__all__ = ["BackgroundLoader", "BackgroundWriter", "DataManager"]

import logging
import os
from typing import Callable, Iterable, List, Tuple
from uuid import UUID

import numpy as np
from PySide2.QtCore import QObject, QStandardPaths, Qt, QThread, Signal
from PySide2.QtWidgets import QFileDialog, QMessageBox, QWidget

from QGrain.algorithms import DistributionType
from QGrain.models.DataLayoutSetting import *
from QGrain.models.DataLoader import *
from QGrain.models.DataWriter import *
from QGrain.models.FittingResult import *
from QGrain.models.SampleData import *
from QGrain.models.SampleDataset import *
from QGrain.resolvers.HeadlessResolver import FittingTask


class BackgroundLoader(QObject):
    sigWorkSucceeded = Signal(SampleDataset)
    sigWorkFailed = Signal(str, Exception)

    def __init__(self):
        super().__init__()
        self.actual_loader = DataLoader()

    def on_work_started(self, filename: str, file_type: FileType,
                        layout: DataLayoutSetting):
        try:
            dataset = self.actual_loader.try_load_data(filename, file_type, layout)
            self.sigWorkSucceeded.emit(dataset)
        except Exception as e:
            self.sigWorkFailed.emit(filename, e)


class BackgroundWriter(QObject):
    sigWorkSucceeded = Signal()
    sigWorkFailed = Signal(str, Exception)

    def __init__(self):
        super().__init__()
        self.actual_writer = DataWriter()

    def on_work_started(self, filename: str, file_type: FileType,
                        results: List[FittingResult], draw_charts: bool):
        try:
            self.actual_writer.try_save_data(filename, file_type, results, draw_charts)
            self.sigWorkSucceeded.emit()
        except Exception as e:
            self.sigWorkFailed.emit(filename, e)


class DataManager(QObject):
    sigLoadingStarted = Signal(str, FileType, DataLayoutSetting)
    sigSavingStarted = Signal(str, FileType, list, bool)
    sigDataLoaded = Signal(SampleDataset)
    sigTargetDataChanged = Signal(SampleData)
    sigDataRecorded = Signal(list) # List[FittingResult]
    sigShowFittingResult = Signal(FittingResult)
    logger = logging.getLogger("root.data.DataManager")
    gui_logger = logging.getLogger("GUI")

    def __init__(self, host_widget: QWidget):
        super().__init__()
        # to attach msg boxed on this widget
        self.host_widget = host_widget
        # data
        self.dataset = None  # type: SampleDataset
        self.current_fitting_result = None  # type: FittingResult
        self.records = []  # type: List[FittingResult]
        # loader
        self.loader = BackgroundLoader()
        self.loading_thread = QThread()
        self.loader.moveToThread(self.loading_thread)
        self.sigLoadingStarted.connect(self.loader.on_work_started)
        self.loader.sigWorkSucceeded.connect(self.on_loading_succeeded)
        self.loader.sigWorkFailed.connect(self.on_loading_failed)
        # writer
        self.writer = BackgroundWriter()
        self.saving_thread = QThread()
        self.writer.moveToThread(self.saving_thread)
        self.sigSavingStarted.connect(self.writer.on_work_started)
        self.writer.sigWorkSucceeded.connect(self.on_saving_succeeded)
        self.writer.sigWorkFailed.connect(self.on_saving_failed)
        # settings
        self.data_layout_setting = DataLayoutSetting()
        self.draw_charts_flag = True
        self.auto_record_flag = True

        self.file_dialog = QFileDialog(self.host_widget)
        self.msg_box = QMessageBox(self.host_widget)
        self.msg_box.setWindowFlags(Qt.Drawer)
        self.retry_msg_box = QMessageBox(self.host_widget)
        self.retry_msg_box.addButton(QMessageBox.StandardButton.Retry)
        self.retry_msg_box.addButton(QMessageBox.StandardButton.Ok)
        self.retry_msg_box.setDefaultButton(QMessageBox.StandardButton.Retry)
        self.retry_msg_box.setWindowFlags(Qt.Drawer)
        self.record_msg_box = QMessageBox(self.host_widget)
        self.record_msg_box.addButton(QMessageBox.StandardButton.Discard)
        self.record_msg_box.addButton(QMessageBox.StandardButton.Ok)
        self.record_msg_box.setDefaultButton(QMessageBox.StandardButton.Discard)
        self.record_msg_box.setWindowFlags(Qt.Drawer)

    def show_message(self, title: str, message: str):
        self.msg_box.setWindowTitle(title)
        self.msg_box.setText(message)
        self.msg_box.exec_()

    def show_info(self, message: str):
        self.show_message(self.tr("Info"), message)

    def show_warning(self, message: str):
        self.show_message(self.tr("Warning"), message)

    def show_error(self, message: str):
        self.show_message(self.tr("Error"), message)

    def ask_retry(self, message: str, func: Callable):
        self.retry_msg_box.setWindowTitle(self.tr("Error"))
        self.retry_msg_box.setText(message)
        result = self.retry_msg_box.exec_()
        if result == QMessageBox.Retry:
            func()

    def load_data(self):
        # NOTE:
        # if don't assign the initial directory,
        # there is an about 5s delay while first calling `getOpenFileName` func
        # if there is a unreachable network disk
        init_path = QStandardPaths.standardLocations(QStandardPaths.DesktopLocation)[0]
        filename, type_str = self.file_dialog.getOpenFileName(
            self.host_widget, self.tr("Select Data File"),
            init_path, "Excel (*.xlsx);;97-2003 Excel (*.xls);;CSV (*.csv)")
        if filename is None or filename == "":
            self.logger.info("The user did not select a file, ignored.")
            return
        # get the file type info
        if ".xlsx" in type_str:
            file_type = FileType.XLSX
        elif "97-2003" in type_str:
            file_type = FileType.XLS
        elif ".csv" in type_str:
            file_type = FileType.CSV
        else:
            raise NotImplementedError(type_str)

        self.logger.info("Try to load data, selected data file is [%s].", filename)
        self.sigLoadingStarted.emit(filename, file_type, self.data_layout_setting)

    def on_loading_succeeded(self, dataset: SampleDataset):
        if self.dataset is None:
            self.dataset = dataset
        else:
            try:
                self.dataset.combine(dataset)
            except ClassesNotMatchError:
                self.logger.exception("The new batch of data's classes are not equal to the existing's.", stack_info=True)
                self.ask_retry(self.tr("The new batch of data's classes are not equal to the existing's."), self.load_data)
                return
        self.sigDataLoaded.emit(self.dataset)
        self.logger.info("Data has been loaded.")
        self.show_info(self.tr("Data has been loaded."))

    def on_loading_failed(self, filename: str, exception: Exception):
        try:
            raise exception
        # handle the exceptions
        except PermissionError:
            self.logger.exception("Can not access the selected file due to permission issue. Filename: [%s].", filename, stack_info=True)
            self.ask_retry(self.tr("Can not access the selected file due to permission issue, please check whether it's opened by another program."), self.load_data)
        except FileNotFoundError:
            self.logger.exception("The selected file does not exist. Maybe it's removed by others. Filename: [%s].", filename, stack_info=True)
            self.ask_retry(self.tr("The selected file does not exist now, please check whether it's removed."), self.load_data)
        except XLRDError:
            self.logger.exception("A exception raised while reading Excel file. Filename: [%s].", filename, stack_info=True)
            self.ask_retry(self.tr("Can not read data from this Excel file."), self.load_data)
        except CSVEncodingError:
            self.logger.exception("The encoding of this CSV file is not utf-8. Filename: [%s].", filename, stack_info=True)
            self.ask_retry(self.tr("Please make sure the encoding of CSV file is utf-8."), self.load_data)
        except ValueNotNumberError:
            self.logger.exception("Can not convert the value to real number. Data layout setting: [%s].", self.data_layout_setting, stack_info=True)
            self.ask_retry(self.tr("The value can not be converted to real number.\nPlease make sure the data layout setting matchs your data file."), self.load_data)
        except ClassesNotIncrementalError:
            self.logger.exception("The classes values are not incremental. Filename: [%s]. Data layout setting: [%s].", filename, self.data_layout_setting, stack_info=True)
            self.ask_retry(self.tr("The array of grain size classes is not incremental.\nPlease make sure the data file is correct and the data layout setting matchs the file."), self.load_data)
        except ClassesNotMatchError:
            self.logger.exception("The new batch of data's classes are not equal to the existing's.", stack_info=True)
            self.ask_retry(self.tr("The new batch of data's classes are not equal to the existing's."), self.load_data)
        except NaNError:
            self.logger.exception("There is a NaN value in this data file.", stack_info=True)
            self.ask_retry(self.tr("There is a NaN value in this data file."), self.load_data)
        except ArrayEmptyError:
            self.logger.exception("There is an array that is empty.", stack_info=True)
            self.ask_retry(self.tr("There is an array that is empty.\nIf you are sure that the CSV file and layout setting are correct, please check the dialect of CSV is excel, NOT excel-tab."), self.load_data)
        except SampleNameEmptyError:
            self.logger.exception("At leaset one sample name is empty.", stack_info=True)
            self.ask_retry(self.tr("At leaset one sample name is empty. Please give each sample a unique name to identify."), self.load_data)
        except DistributionSumError:
            self.logger.exception("The sum of distribution array is not equal to 1 or 100.", stack_info=True)
            self.ask_retry(self.tr("The sum of distribution array is not equal to 1 or 100. Please make sure the data file you selected is grain size distribution data."), self.load_data)
        except Exception:
            self.logger.exception("Unknown exception raised.", stack_info=True)
            self.ask_retry(self.tr("Unknown exception raised. See the log for more details."), self.load_data)

    def on_focus_sample_changed(self, index: int):
        if self.dataset is None:
            self.logger.info("Grain size data is still None, ignored.")
            return
        self.sigTargetDataChanged.emit(self.dataset.samples[index])
        self.logger.debug("Focus sample data changed, the data has been emitted.")

    def on_fitting_suceeded(self, result: FittingResult):
        self.logger.info("Epoch for sample [%s] has finished, mean squared error is [%E].", result.name, result.mean_squared_error)
        self.current_fitting_result = result
        if self.auto_record_flag:
            if result.has_invalid_value:
                self.record_msg_box.setWindowTitle(self.tr("Warning"))
                self.record_msg_box.setText(self.tr("The fitting result may be invalid."))
                exec_result = self.record_msg_box.exec_()
                if exec_result == QMessageBox.Discard:
                    self.logger.info("Fitting result of sample [%s] was discarded by user.", result.name)
                    return

            self.record_current_data()
            self.sigShowFittingResult.emit(result)

    def on_multiprocessing_task_finished(self, succeeded_results: Iterable[FittingResult], failed_tasks: Iterable[FittingTask]):
        for fitting_result in succeeded_results:
            if fitting_result.has_invalid_value:
                self.logger.warning("There is invalid value in the fitting result of sample [%s].", fitting_result.name)
                self.gui_logger.warning(self.tr("There is invalid value in the fitting result of sample [%s]."), fitting_result.name)
        self.records.extend(succeeded_results)
        self.sigDataRecorded.emit(succeeded_results)
        for failed_task in failed_tasks:
            self.logger.warning("Fitting task of sample [%s] failed.", failed_task.sample.name)
            self.gui_logger.warning(self.tr("Fitting task of sample [%s] failed."), failed_task.sample.name)

    def on_settings_changed(self, kwargs: dict):
        for key, value in kwargs.items():
            if key == "layout":
                self.data_layout_setting = value
            elif key == "draw_charts":
                self.draw_charts_flag = value
            elif key == "auto_record":
                self.auto_record_flag = value
            else:
                raise NotImplementedError(key)
            self.logger.info("Setting [%s] have been changed to [%s].", key, value)

    def record_current_data(self):
        if self.current_fitting_result is None:
            self.logger.info("There is no fitting result to record, ignored.")
            self.show_warning(self.tr("There is no fitting result to record."))
            return
        self.records.append(self.current_fitting_result)
        self.sigDataRecorded.emit([self.current_fitting_result])

    def remove_data(self, uuids_and_names: Iterable[Tuple[UUID, str]]):
        for uuid, name in uuids_and_names:
            for i, data in enumerate(self.records):
                if uuid == data.uuid:
                    assert name == data.name
                    self.records.pop(i)
                    self.logger.info("Record of sample [%s] has been removed.", name)
                    break

    def save_data(self):
        filename, type_str = self.file_dialog.getSaveFileName(None, self.tr("Save Recorded Data"), None, "Excel (*.xlsx);;97-2003 Excel (*.xls);;CSV (*.csv)")
        if filename is None or filename == "":
            self.logger.info("The path is None or empty, ignored.")
            return
        if os.path.exists(filename):
            self.logger.warning("This file has existed and will be replaced. Filename: %s.", filename)
        self.logger.info("File path to save is [%s].", filename)
        if ".xlsx" in type_str:
            file_type = FileType.XLSX
        elif "97-2003" in type_str:
            file_type = FileType.XLS
        elif ".csv" in type_str:
            file_type = FileType.CSV
        else:
            raise NotImplementedError(type_str)
        self.logger.info("Selected file type is [%s].", file_type)
        self.sigSavingStarted.emit(filename, file_type, self.records, self.draw_charts_flag)

    def on_saving_succeeded(self):
            self.logger.info("File saved.")
            self.show_info(self.tr("The data has been saved to the file."))

    def on_saving_failed(self, filename: str, exception: Exception):
        try:
            raise exception
        except PermissionError:
            self.logger.exception("Can not access the selected file due to permission issue. Filename: [%s].", filename, stack_info=True)
            self.ask_retry(self.tr("Can not access the selected file due to permission issue, please check whether it's opened by another program."), self.save_data)
        except Exception:
            self.logger.exception("Unknown exception raised.", stack_info=True)
            self.ask_retry(self.tr("Unknown exception raised. See the log for more details."), self.save_data)

    def setup_all(self):
        self.loading_thread.start()
        self.saving_thread.start()

    def cleanup_all(self):
        self.loading_thread.terminate()
        self.saving_thread.terminate()



import os
import random

import numpy as np
from scipy.interpolate import interp1d
from scipy.stats import norm

from QGrain import QGRAIN_ROOT_PATH


def get_scale_ratio(max_ratio):
    assert max_ratio > 0.0
    return (random.random() - 0.5) / 0.5 * max_ratio  + 1.0

def get_random_sample(classes, params, max_scale_ratios, noise=0.005):
    assert len(params) == len(max_scale_ratios)
    bin_numbers = np.array(range(len(classes))) + 1
    converter = interp1d(classes, bin_numbers)
    components = []
    # get random params
    actual_params = []
    for i in range(len(params)):
        assert len(params[i]) == 3
        assert len(max_scale_ratios[i]) == 3
        component_prams = []
        for j in range(3):
            param = params[i][j] * get_scale_ratio(max_scale_ratios[i][j])
            component_prams.append(param)
        actual_params.append(component_prams)
    # standardize the fractions to make its sum equals 1
    sum_of_fractions = sum(component_prams[2] for component_prams in actual_params)
    for component_prams in actual_params:
        component_prams[2] /= sum_of_fractions
    # generate distributions
    distribution = np.zeros_like(classes)
    for mean, std, fraction in actual_params:
        component_distribution = norm.pdf(bin_numbers, converter(mean), std) * fraction
        components.append(component_distribution)
        np.add(distribution, component_distribution, out=distribution)
    # add noise
    noise_array = np.random.normal(0.0, noise, len(classes)) + 1.0
    np.multiply(distribution, noise_array, out=distribution)
    # standardize
    np.divide(distribution, np.sum(distribution), out=distribution)
    return distribution, components, actual_params


if __name__ == "__main__":
    from PySide2.QtCore import Qt, QPointF, QTimer
    from PySide2.QtCharts import QtCharts
    from PySide2.QtWidgets import QApplication
    from PySide2.QtGui import QPainter, QFont, QIcon
    import sys

    sample_classes = np.logspace(0, 5, 101) * 0.02
    # sample_params = ((1.1, 6, 0.1), (8.0, 9.0, 0.5), (30.0, 4.5, 0.4))
    # sample_max_scale_ratios = ((0.2, 0.2, 0.4), (0.4, 0.4, 0.6), (0.4, 0.4, 0.6))
    sample_params = ((1.1, 6, 0.1), (8.0, 9.0, 0.2), (30.0, 4.5, 0.4), (300, 6.5, 0.3))
    sample_max_scale_ratios = ((0.2, 0.2, 0.4), (0.2, 0.2, 0.4), (0.4, 0.4, 0.8), (0.3, 0.3, 0.6))

    app = QApplication(sys.argv)
    app.setWindowIcon(QIcon(os.path.join(QGRAIN_ROOT_PATH, "settings", "icons", "icon.png")))
    chart = QtCharts.QChart()
    chart.setBackgroundVisible(False)
    # chart.setAnimationOptions(QtCharts.QChart.SeriesAnimations)
    chart.setTitle("Artificial samples")
    chart.setTitleFont(QFont("Times New Roman", 14))
    axis_x = QtCharts.QLogValueAxis()
    axis_x.setTitleText("Grain size (μm)")
    axis_x.setTitleFont(QFont("Times New Roman", 12))
    axis_y = QtCharts.QValueAxis()
    axis_y.setLabelFormat("%0.3f")
    axis_y.setTitleText("Probability density")
    axis_y.setTitleFont(QFont("Times New Roman", 12))
    chart.addAxis(axis_x, Qt.AlignBottom)
    chart.addAxis(axis_y, Qt.AlignLeft)

    sum_series = QtCharts.QLineSeries()
    sum_series.setName("Sum")
    chart.addSeries(sum_series)
    sum_series.attachAxis(axis_x)
    sum_series.attachAxis(axis_y)
    components_series = []
    for i in range(len(sample_params)):
        series = QtCharts.QLineSeries()
        series.setName("C" + str(i+1))
        chart.addSeries(series)
        series.attachAxis(axis_x)
        series.attachAxis(axis_y)
        components_series.append(series)

    def to_points(x, y):
        return [QPointF(x_value, y_value) for x_value, y_value in zip(x, y)]

    def update():
        # prepare data
        distribution, components, _ = get_random_sample(sample_classes, sample_params, sample_max_scale_ratios)

        sum_series.replace(to_points(sample_classes, distribution))
        for series, component in zip(components_series, components):
            series.replace(to_points(sample_classes, component))
        axis_y.setRange(0.0, round(np.max(distribution)*1.2, 2))

    axis_x.setRange(sample_classes[0], sample_classes[-1])
    update()
    view = QtCharts.QChartView()
    view.setChart(chart)
    view.setRenderHint(QPainter.Antialiasing)
    view.setWindowTitle("Artificial Sample Generator")
    view.setGeometry(200, 200, 600, 500)
    view.show()

    timer = QTimer()
    timer.timeout.connect(update)
    # timer.start(2000)
    timer.start(100)

    sys.exit(app.exec_())



__all__ = ["AllDistributionCanvas"]

import logging

import numpy as np
from PySide2.QtCharts import QtCharts
from PySide2.QtCore import Qt
from PySide2.QtGui import QColor, QPen

from QGrain.models.SampleData import SampleData
from QGrain.models.SampleDataset import SampleDataset
from QGrain.ui.Canvas import Canvas


class AllDistributionCanvas(Canvas):
    logger = logging.getLogger("root.ui.AllDistributionCanvas")
    gui_logger = logging.getLogger("GUI")

    def __init__(self, parent=None, is_dark=True):
        super().__init__("AllDistributionCanvas", parent=parent)
        self.set_theme_mode(is_dark)
        self.init_chart()
        self.setup_chart_style()
        self.chart.legend().hide()

    def init_chart(self):
        # init axes
        self.axis_x = QtCharts.QLogValueAxis()
        self.axis_x.setBase(10.0)
        self.axis_x.setMinorTickCount(-1)
        self.chart.addAxis(self.axis_x, Qt.AlignBottom)
        self.axis_y = QtCharts.QValueAxis()
        self.chart.addAxis(self.axis_y, Qt.AlignLeft)
        # set title
        self.chart.setTitle(self.tr("All Distribution Canvas"))
        # set labels
        self.axis_x.setTitleText(self.tr("Grain size")+" (μm)")
        self.axis_y.setTitleText(self.tr("Probability Density"))

        self.show_demo(self.axis_x, self.axis_y, x_log=True)

    def on_data_loaded(self, dataset: SampleDataset):
        self.stop_demo()
        self.chart.removeAllSeries()
        max_y = -1.0
        for sample in dataset.samples:
            series = QtCharts.QLineSeries()
            series.setName(sample.name)
            series.replace(self.to_points(sample.classes, sample.distribution))
            self.chart.addSeries(series)
            pen = QPen(QColor(series.pen().color()),
                       1.0)
            series.setPen(pen)
            series.attachAxis(self.axis_x)
            series.attachAxis(self.axis_y)
            current_max_y = np.max(sample.distribution)
            if current_max_y > max_y:
                max_y = current_max_y
        self.axis_x.setRange(dataset.classes[0], dataset.classes[-1])
        self.axis_y.setRange(0.0, round(max_y*1.2, 2))


if __name__ == "__main__":
    import sys
    from PySide2.QtWidgets import QApplication
    app = QApplication(sys.argv)
    canvas = AllDistributionCanvas(is_dark=False)
    canvas.chart.legend().hide()
    canvas.show()
    sys.exit(app.exec_())



__all__ = ["ClassesNotIncrementalError",
           "ClassesNotMatchError",
           "NaNError",
           "ArrayEmptyError",
           "SampleNameEmptyError",
           "DistributionSumError",
           "SampleDataset"]

from typing import Callable, Dict, Iterable, List, Tuple
from uuid import UUID

import numpy as np

from QGrain.models.SampleData import SampleData


class ClassesNotIncrementalError(Exception):
    """Raises while the array of grain size classes is not incremental.

    It's just an ASSUMPTION for the convenience of coding.
    """
    pass


class ClassesNotMatchError(Exception):
    """Raises while the grain size classes of a new batch of sample data are not equal to the existing classes."""
    pass


class NaNError(Exception):
    """Raises while there is at least one NaN value in the array."""
    pass


class ArrayEmptyError(Exception):
    """Raises while the length of array is 0."""
    pass


class SampleNameEmptyError(Exception):
    """Raises while the name of sample is empty."""
    pass


class DistributionSumError(Exception):
    """Raises while the sum of distribution array is not equal to 1 or 100."""
    pass


class SampleDataset:
    def __init__(self):
        self.__classes = None
        self.__samples = []

    @property
    def data_count(self) -> int:
        return len(self.__samples)

    @property
    def has_data(self) -> bool:
        if self.data_count == 0:
            return False
        else:
            return True

    @property
    def classes(self) -> np.ndarray:
        return self.__classes

    @property
    def samples(self) -> Iterable[SampleData]:
        return [sample for sample in self.__samples]

    def is_incremental(self, nums: np.ndarray) -> bool:
        """Returns `True` while the array is incremental.
        This method is used to validate the array of grain size classes."""
        for i in range(1, len(nums)):
            if nums[i] <= nums[i-1]:
                return False
        return True

    def validate_classes(self, classes: np.ndarray):
        assert classes is not None
        assert type(classes) == np.ndarray
        assert classes.dtype == np.float64
        if len(classes) == 0:
            raise ArrayEmptyError()
        if np.any(np.isnan(classes)):
            raise NaNError(classes)
        # may raise when users select a wrong data file
        if not self.is_incremental(classes):
            raise ClassesNotIncrementalError(classes)

    def validate_sample_name(self, sample_name: str):
        assert sample_name is not None
        assert type(sample_name) == str
        if sample_name == "":
            raise SampleNameEmptyError()

    def validate_distribution(self, distribution: np.ndarray):
        assert distribution is not None
        assert type(distribution) == np.ndarray
        assert distribution.dtype == np.float64
        if len(distribution) == 0:
            raise ArrayEmptyError()
        if np.any(np.isnan(distribution)):
            raise NaNError(distribution)
        # check the sum is close to 1 or 100 (someone may use percentage)
        s = np.sum(distribution)
        if (s > 0.99 and s < 1.01) or (s > 99 and s < 101):
            np.true_divide(distribution, s, out=distribution)
        else:
            raise DistributionSumError(distribution)

    def add_batch(self, classes: np.ndarray, names: Iterable[str],
                  distributions: Iterable[np.ndarray]):
        self.validate_classes(classes)
        # If there is data already, it's necessary to check the consistency of classes.
        if self.has_data:
            equal_res = np.equal(self.classes, classes)
            if not np.all(equal_res):
                raise ClassesNotMatchError(self.classes, classes)

        assert len(names) == len(distributions)
        # use temp to implement roll-back
        temp = []
        for name, distribution in zip(names, distributions):
            assert len(distribution) == len(classes)
            self.validate_sample_name(name)
            self.validate_distribution(distribution)
            sample = SampleData(name, classes, distribution)
            temp.append(sample)
        # if no exception raised, the codes below will be executed
        self.__samples.extend(temp)
        self.__classes = classes

    def combine(self, another_dataset):
        equal_res = np.equal(another_dataset.classes, self.classes)
        if np.all(equal_res):
            self.__samples.extend(another_dataset.samples)
        else:
            raise ClassesNotMatchError(self.classes, another_dataset.classes)

    def get_sample_by_id(self, uuid: UUID):
        for sample in self.samples:
            if sample.uuid == uuid:
                return sample
        # TODO: raise custom exception
        raise ValueError("There is no sample with this id.", uuid)

    def remove_sample_by_id(self, uuid: UUID):
        index_to_remove = None
        for index, sample in enumerate(self.samples):
            if sample.uuid == uuid:
                index_to_remove = index
                break
        if index_to_remove is not None:
            removed_sample = self.__samples.pop(index_to_remove)
            return removed_sample
        else:
            raise ValueError("There is no sample with this id.", uuid)



__all__ = ["SampleTag", "SampleData"]


from enum import Enum, unique
from uuid import uuid4, UUID
from typing import Callable, Dict, Iterable, List, Tuple

import numpy as np

@unique
class SampleTag(Enum):
    Default = 1
    Ignored = 2


class SampleData:
    """This class is used to store all the necessary data of each sample which can be fitted.

    All its attributes have been packed with read-only properties to avoid the modification by mistake.

    Note: Do not call `ctor` to create instances because it won't check the validity of input arguments.
    You can use the `add_batch` func of class `SampleDataset` instead.

    Attributes:
        uuid: A `uuid.UUID` object.
        name: A human readable `str`.
        classes: A `numpy.ndarray` to represent the grain size classes.
        distribution: A `numpy.ndarray` to represent the distribution data of this sample.
        tag: A `SampleTag` enum to indicate that how to handle this sample. Use `reset_tag`, `ignore` to control the tag.
    """
    def __init__(self, name: str, classes: np.ndarray, distribution: np.ndarray):
        self.__uuid = uuid4()
        self.__name = name
        self.__classes = classes
        self.__distribution = distribution
        self.__tag = SampleTag.Default

    @property
    def uuid(self) -> UUID:
        return self.__uuid

    @property
    def name(self) -> str:
        return self.__name

    @property
    def classes(self) -> np.ndarray:
        return self.__classes

    @property
    def distribution(self) -> np.ndarray:
        return self.__distribution

    @property
    def tag(self) -> SampleTag:
        return self.__tag

    def reset_tag(self):
        self.__tag = SampleTag.Default

    def ignore(self):
        self.__tag = SampleTag.Ignored



__all__ = ["DataLayoutError", "DataLayoutSetting"]

class DataLayoutError(Exception):
    """Raises while the data layout settings are invalid."""
    pass


class DataLayoutSetting:
    """
    The class to represent the layout setting of raw data file.
    All types of raw data files can be regarded as the table(s) that contains rows and columns.
    For the grain size distribution files, it should be the following format:
        * The first valid row should be the headers (i.e. the classes of grain size).
        * The following valid rows should be the distributions of samples under the grain size classes.
        * The first valid column shoud be the name (i.e. id) of samples.
    To make it more flexible, we use this setting to control the data loader.
    """
    def __init__(self, classes_row=0, sample_name_column=0,
                 distribution_start_row=1, distribution_start_column=1):
        # make sure the types are int in other codes
        assert type(classes_row) == int
        assert type(sample_name_column) == int
        assert type(distribution_start_row) == int
        assert type(distribution_start_column) == int

        # handle these errors at front-end
        if classes_row < 0 or \
            sample_name_column < 0 or \
            distribution_start_row < 0 or \
            distribution_start_column < 0:
            raise DataLayoutError("Row or column index must be non-negative.")
        if classes_row >= distribution_start_row:
            raise DataLayoutError("The start row index of distribution data must be greater than the row index of classes.")
        if sample_name_column >= distribution_start_column:
            raise DataLayoutError("The start column index of distribution data must be greater than the column index of sample names.")

        self.__classes_row = classes_row
        self.__sample_name_column = sample_name_column
        self.__distribution_start_row = distribution_start_row
        self.__distribution_start_column = distribution_start_column

    @property
    def classes_row(self) -> int:
        return self.__classes_row

    @property
    def sample_name_column(self) -> int:
        return self.__sample_name_column

    @property
    def distribution_start_row(self) -> int:
        return self.__distribution_start_row

    @property
    def distribution_start_column(self) -> int:
        return self.__distribution_start_column

    def __str__(self):
        return "Class Row Index: {0}, Sample Name Column: {1}, Distribution Start Row: {2}, Distribution Start Column: {3}.".format(
            self.classes_row, self.sample_name_column, self.distribution_start_row, self.distribution_start_column)



__all__ = ["DistributionCanvas"]

import logging
from typing import List

import cv2
import numpy as np
from PIL import Image
from PySide2.QtCharts import QtCharts
from PySide2.QtCore import (QCoreApplication, QEventLoop, QMutex,
                            QStandardPaths, Qt, QTimer, Signal)
from PySide2.QtGui import QColor, QPen
from PySide2.QtWidgets import QFileDialog

from QGrain.models.FittingResult import FittingResult
from QGrain.models.SampleData import SampleData
from QGrain.ui.Canvas import Canvas
from QGrain.ui.InfiniteLine import InfiniteLine


class DistributionCanvas(Canvas):
    sigExpectedMeanValueChanged = Signal(tuple)
    logger = logging.getLogger("root.ui.DistributionCanvas")
    gui_logger = logging.getLogger("GUI")

    def __init__(self, parent=None, is_dark=True):
        super().__init__("DistributionCanvas", parent=parent)
        self.set_theme_mode(is_dark)
        self.init_chart()
        self.setup_chart_style()
        self.chart.legend().detachFromChart()
        self.chart.legend().setPos(100.0, 60.0)
        self.__infinite_line_mutex = QMutex()
        self.__iteration_mutex = QMutex()
        self.__iteration_timer = QTimer()
        # used to check if it's necessary to update the component series and lines
        self.__current_component_number = None
        self.video_format_options = {self.tr("MP4 (*.mp4)"): "mp4v",
                                     self.tr("AVI Motion-JPEG (*.avi)"): "MJPG",
                                     self.tr("AVI MPEG-4.2 (*.avi)"): "MP42",
                                     self.tr("AVI MPEG-4.3 (*.avi)"): "DIV3",
                                     self.tr("AVI MPEG-4 (*.avi)"): "DIVX"}
        self.file_dialog = QFileDialog(self)

    def init_chart(self):
        # init axes
        self.axis_x = QtCharts.QLogValueAxis()
        self.axis_x.setBase(10.0)
        self.axis_x.setMinorTickCount(-1)
        self.chart.addAxis(self.axis_x, Qt.AlignBottom)
        self.axis_y = QtCharts.QValueAxis()
        self.chart.addAxis(self.axis_y, Qt.AlignLeft)

        # init two const series
        self.target_series = QtCharts.QScatterSeries()
        self.target_series.setName(self.tr("Target"))
        self.target_series.setMarkerSize(5.0)
        self.target_series.setPen(QPen(QColor(255, 255, 255, 0)))
        self.chart.addSeries(self.target_series)
        self.fitted_series = QtCharts.QLineSeries()
        self.fitted_series.setName(self.tr("Fitted"))
        self.chart.addSeries(self.fitted_series)
        # attach series to axes
        self.target_series.attachAxis(self.axis_x)
        self.target_series.attachAxis(self.axis_y)
        self.fitted_series.attachAxis(self.axis_x)
        self.fitted_series.attachAxis(self.axis_y)

        # set title
        self.chart.setTitle(self.tr("Distribution Canvas"))
        # set labels
        self.axis_x.setTitleText(self.tr("Grain size")+" (μm)")
        self.axis_y.setTitleText(self.tr("Probability Density"))

        self.show_demo(self.axis_x, self.axis_y, x_log=True)
        self.component_series = [] # type: List[QtChart.QLineSeries]
        self.component_infinite_lines = [] # type: List[InfiniteLine]
        self.observe_iteration_tag = False

    def on_component_number_changed(self, component_number: int):
        # use assert because the error must be handled in other place
        assert isinstance(component_number, int)
        assert component_number > 0
        # clear
        for series in self.component_series:
            self.chart.removeSeries(series)
        for line in self.component_infinite_lines:
            line.disconnect_from_chart()
        self.component_series.clear()
        self.component_infinite_lines.clear()
        # add
        for i in range(component_number):
            component_name = "C{0}".format(i+1)
            # series
            series = QtCharts.QLineSeries()
            series.setName(component_name)
            self.chart.addSeries(series)
            series.attachAxis(self.axis_x)
            series.attachAxis(self.axis_y)
            # line
            line = InfiniteLine(is_horizontal=False, chart=self.chart,
                                pen=series.pen(), callback=self.on_infinite_line_moved)
            self.component_series.append(series)
            self.component_infinite_lines.append(line)
        # update the size of legend
        self.chart.legend().setMinimumSize(150.0, 30*(2+component_number))
        self.__current_component_number = component_number

    def on_observe_iteration_changed(self, value: bool):
        self.observe_iteration_tag = value

    def show_target_distribution(self, sample: SampleData):
        # necessary to stop
        self.stop_demo()
        # update the title of canvas
        self.chart.setTitle(sample.name)
        self.target_series.replace(self.to_points(sample.classes, sample.distribution))
        self.fitted_series.clear()
        for series in self.component_series:
            series.clear()
        for line in self.component_infinite_lines:
            line.value = 1.0
        self.axis_x.setRange(sample.classes[0], sample.classes[-1])
        self.axis_y.setRange(0.0, round(np.max(sample.distribution)*1.2, 2))

    def __update_components(self, result: FittingResult):
        # update fitted series
        self.fitted_series.replace(self.to_points(result.real_x, result.fitted_y))
        # update component series
        for component_index, (component, series, line) in enumerate(zip(
                result.components,
                self.component_series,
                self.component_infinite_lines)):
            series.replace(self.to_points(result.real_x, component.component_y))
            # display the median (modal size) and fraction for users
            component_name = "C{0}".format(component_index+1)
            series.setName(component_name + " ({0:.1f} μm, {1:.1%})".format(component.median, component.fraction))
            if np.isnan(component.mean) or np.isinf(component.mean):
                # if mean is invalid, set the position of this line to initial position
                line.value = 1.0
            else:
                line.value = component.mean
        self.axis_x.setRange(result.real_x[0], result.real_x[-1])
        self.axis_y.setRange(0.0, round(np.max(result.target_y)*1.2, 2))

    def show_fitting_result(self, result: FittingResult):
        # necessary to stop
        self.stop_demo()
        success = self.__iteration_mutex.tryLock()
        if not success:
            return
        # update target series
        self.target_series.replace(self.to_points(result.real_x, result.target_y))
        # check the component number
        if self.__current_component_number != result.component_number:
            self.on_component_number_changed(result.component_number)

        generator = result.history
        current_iteration = 0

        def closure():
            try:
                nonlocal current_iteration
                result_in_process = next(generator)
                self.chart.setTitle(("{0} "+self.tr("Iteration")+" ({1})").format(
                    result.name, current_iteration))
                self.__update_components(result_in_process)
                current_iteration += 1
            except StopIteration:
                self.__iteration_timer.stop()
                self.__iteration_timer.timeout.disconnect(closure)
                self.chart.setTitle(result.name)
                self.__update_components(result)
                self.__iteration_mutex.unlock()

        if self.observe_iteration_tag:
            self.__iteration_timer.timeout.connect(closure)
            self.__iteration_timer.start(1000//30)
        else:
            self.chart.setTitle(result.name)
            self.__update_components(result)
            self.__iteration_mutex.unlock()

    def generate_video(self, result: FittingResult):
        # necessary to stop
        self.stop_demo()
        # update target series
        self.target_series.replace(self.to_points(result.real_x, result.target_y))
        # check the component number
        if self.__current_component_number != result.component_number:
            self.on_component_number_changed(result.component_number)
        width = self.export_dialog.width
        height = self.export_dialog.height
        pixel_ratio = self.export_dialog.pixel_ratio
        video_size = (int(width*pixel_ratio), int(height*pixel_ratio))
        fps = 30.0
        desktop_path = QStandardPaths.standardLocations(QStandardPaths.DesktopLocation)[0]
        filename, format_name = self.file_dialog.getSaveFileName(
            self, self.tr("Select Filename"),
            desktop_path, ";;".join([format_name for format_name, fourcc in self.video_format_options.items()]))
        video = cv2.VideoWriter(filename, cv2.VideoWriter_fourcc(*self.video_format_options[format_name]), fps, video_size)
        for current_iteration, result_in_process in enumerate(result.history):
            self.chart.setTitle(("{0} "+self.tr("Iteration")+" ({1})").format(
                    result.name, current_iteration))
            self.__update_components(result_in_process)
            QCoreApplication.processEvents(QEventLoop.ExcludeUserInputEvents)
            pixmap = self.get_pixmap(width=width, height=height, pixel_ratio=pixel_ratio)
            mat = cv2.cvtColor(np.asarray(Image.fromqpixmap(pixmap)), cv2.COLOR_RGB2BGR)
            video.write(mat)

        video.release()

    def on_infinite_line_moved(self):
        # this method will be called by another object directly
        # use lock to keep safety
        self.__infinite_line_mutex.lock()
        expected_mean_values = tuple(sorted([line.value for line in self.component_infinite_lines]))
        self.sigExpectedMeanValueChanged.emit(expected_mean_values)
        self.__infinite_line_mutex.unlock()


if __name__ == "__main__":
    import sys
    from PySide2.QtWidgets import QApplication
    app = QApplication(sys.argv)
    canvas = DistributionCanvas(is_dark=False)
    canvas.chart.legend().hide()
    canvas.show()
    sys.exit(app.exec_())



__all__ = ["AboutWindow"]

import os

from PySide2.QtCore import Qt, QUrl
from PySide2.QtWidgets import (QGridLayout, QMainWindow, QTextBrowser,
                               QTextEdit, QWidget)

QGRAIN_ROOT_PATH = os.path.dirname(os.path.dirname(__file__))


class AboutWindow(QMainWindow):
    def __init__(self, parent=None):
        super().__init__(parent=parent)
        self.init_ui()

    def init_ui(self):
        self.central_widget = QWidget(self)
        self.setCentralWidget(self.central_widget)
        self.layout = QGridLayout(self.central_widget)
        self.setWindowTitle(self.tr("About"))
        self.setMinimumSize(600, 400)
        self.text = QTextBrowser()
        self.layout.addWidget(self.text, 0, 0)
        self.setWindowFlags(Qt.Drawer)
        with open(os.path.join(QGRAIN_ROOT_PATH, "settings", "about.md"), mode="r") as text:
            self.text.setMarkdown(text.read())
        self.text.setOpenExternalLinks(True)


if __name__ == "__main__":
    import sys
    from PySide2.QtWidgets import QApplication
    app = QApplication(sys.argv)
    s = AboutWindow()
    s.show()
    app.exec_()



__all__ = ["DataSetting"]

import logging
import os

from PySide2.QtCore import QSettings, Qt, Signal
from PySide2.QtGui import QIcon, QIntValidator, QValidator
from PySide2.QtWidgets import (QCheckBox, QGridLayout, QLabel, QLineEdit,
                               QMessageBox, QWidget)

from QGrain.models.DataLayoutSetting import DataLayoutError, DataLayoutSetting

QGRAIN_ROOT_PATH = os.path.dirname(os.path.dirname(__file__))

class DataSetting(QWidget):
    sigDataSettingChanged = Signal(dict)
    logger = logging.getLogger("root.ui.DataSetting")
    gui_logger = logging.getLogger("GUI")
    def __init__(self):
        super().__init__()
        self.msg_box = QMessageBox(self)
        self.msg_box.setWindowFlags(Qt.Drawer)
        self.settings = QSettings(os.path.join(QGRAIN_ROOT_PATH, "settings", "QGrain.ini"), QSettings.Format.IniFormat)
        self.settings.beginGroup("data")
        self.setAttribute(Qt.WA_StyledBackground, True)
        self.init_ui()

    def init_ui(self):
        self.main_layout = QGridLayout(self)
        self.int_validator = QIntValidator()
        self.int_validator.setBottom(0)

        self.title_label = QLabel(self.tr("Data Settings:"))
        self.title_label.setStyleSheet("QLabel {font: bold;}")
        self.main_layout.addWidget(self.title_label, 0, 0)

        self.class_row_label = QLabel(self.tr("Class Row"))
        self.class_row_label.setToolTip(self.tr("The row index (starts with 0) of grain size classes."))
        self.main_layout.addWidget(self.class_row_label, 1, 0)
        self.class_row_edit = QLineEdit()
        self.class_row_edit.setValidator(self.int_validator)
        self.main_layout.addWidget(self.class_row_edit, 1, 1)

        self.sample_name_col_label = QLabel(self.tr("Sample Name Column"))
        self.sample_name_col_label.setToolTip(self.tr("The column index (starts with 0) of sample names."))
        self.main_layout.addWidget(self.sample_name_col_label, 2, 0)
        self.sample_name_col_edit = QLineEdit()
        self.sample_name_col_edit.setValidator(self.int_validator)
        self.main_layout.addWidget(self.sample_name_col_edit, 2, 1)

        self.distribution_start_row_label = QLabel(self.tr("Distribution Start Row"))
        self.distribution_start_row_label.setToolTip(self.tr("The start row index (starts with 0) of distribution data.\nIt should be greater than the row index of classes."))
        self.main_layout.addWidget(self.distribution_start_row_label, 3, 0)
        self.distribution_start_row_edit = QLineEdit()
        self.distribution_start_row_edit.setValidator(self.int_validator)
        self.main_layout.addWidget(self.distribution_start_row_edit, 3, 1)

        self.distribution_start_col_label = QLabel(self.tr("Distribution Start Column"))
        self.distribution_start_col_label.setToolTip(self.tr("The start column index (starts with 0) of distribution data.\nIt should be greater than the column index of sample name."))
        self.main_layout.addWidget(self.distribution_start_col_label, 4, 0)
        self.distribution_start_col_edit = QLineEdit()
        self.distribution_start_col_edit.setValidator(self.int_validator)
        self.main_layout.addWidget(self.distribution_start_col_edit, 4, 1)

        self.draw_charts_checkbox = QCheckBox(self.tr("Draw Charts"))
        self.draw_charts_checkbox.setChecked(True)
        self.draw_charts_checkbox.setToolTip(self.tr("Whether to draw charts while saving .xlsx file.\nIf the samples are too many, the massive charts will slow the running of Excel heavily."))
        self.main_layout.addWidget(self.draw_charts_checkbox, 5, 0, 1, 2)

        self.class_row_edit.textChanged.connect(self.on_loading_setting_changed)
        self.sample_name_col_edit.textChanged.connect(self.on_loading_setting_changed)
        self.distribution_start_row_edit.textChanged.connect(self.on_loading_setting_changed)
        self.distribution_start_col_edit.textChanged.connect(self.on_loading_setting_changed)

        self.restore()

    def on_loading_setting_changed(self):
        # read settings from ui
        classes_row = self.class_row_edit.text()
        sample_name_column = self.sample_name_col_edit.text()
        distribution_start_row = self.distribution_start_row_edit.text()
        distribution_start_column = self.distribution_start_col_edit.text()
        # emit signal
        try:
            layout = DataLayoutSetting(int(classes_row),
                                       int(sample_name_column),
                                       int(distribution_start_row),
                                       int(distribution_start_column))
            self.sigDataSettingChanged.emit({"layout": layout})
        except DataLayoutError:
            self.logger.exception("The data layout setting is invalid.", stack_info=True)
            self.msg_box.setWindowTitle(self.tr("Error"))
            self.msg_box.setText(self.tr("Invalid data layout setting."))
            self.msg_box.exec_()
            return
        except ValueError:
            return
        self.settings.setValue("classes_row", classes_row)
        self.settings.setValue("sample_name_column", sample_name_column)
        self.settings.setValue("distribution_start_row", distribution_start_row)
        self.settings.setValue("distribution_start_column", distribution_start_column)

    def on_draw_charts_changed(self):
        if self.draw_charts_checkbox.checkState() == Qt.Checked:
            self.settings.setValue("draw_charts", True)
            self.sigDataSettingChanged.emit(dict(draw_charts=True))
        else:
            self.settings.setValue("draw_charts", True)
            self.sigDataSettingChanged.emit(dict(draw_charts=False))

    def restore(self):
        self.class_row_edit.setText(self.settings.value("classes_row", defaultValue="0", type=str))
        self.sample_name_col_edit.setText(self.settings.value("sample_name_column", defaultValue="0", type=str))
        self.distribution_start_row_edit.setText(self.settings.value("distribution_start_row", defaultValue="1", type=str))
        self.distribution_start_col_edit.setText(self.settings.value("distribution_start_column", defaultValue="1", type=str))

        if self.settings.value("draw_charts", defaultValue=True, type=bool):
            self.draw_charts_checkbox.setCheckState(Qt.Checked)
        else:
            self.draw_charts_checkbox.setCheckState(Qt.Unchecked)



__all__ = ["InfiniteLine"]

from typing import Optional, Union, Tuple, Callable

from PySide2.QtCharts import QtCharts
from PySide2.QtCore import QPointF, QRectF, QSizeF, Qt
from PySide2.QtGui import QColor, QPainter, QPen
from PySide2.QtWidgets import (QGraphicsItem, QGraphicsSceneHoverEvent,
                               QGraphicsSceneMouseEvent,
                               QStyleOptionGraphicsItem, QWidget)


class InfiniteLine(QGraphicsItem):
    def __init__(self, is_horizontal: bool, chart: QtCharts.QChart,
                 pen: Optional[QPen] = None, hover_pen: Optional[QPen] = None,
                 callback: Callable = None):
        super().__init__()
        self.__is_horizontal = is_horizontal
        self.__chart = chart
        self.__callback = callback

        self.__pen = QPen(QColor(0x000000), 3.0) if pen is None else pen
        self.__hover_pen = QPen(QColor(0xff0000), 5.0) if hover_pen is None else hover_pen
        self.__current_pen = self.__pen

        self.__value = None
        self.value = sum(self.value_range) / 2
        assert self.__value is not None

        self.setAcceptHoverEvents(True)
        self.connect_to_chart(chart)

    def boundingRect(self) -> QRectF:
        w = self.__current_pen.widthF()
        if self.is_horizontal:
            return QRectF(QPointF(0.0, -w/2), QSizeF(self.length, w*2))
        else:
            return QRectF(QPointF(-w/2, 0.0), QSizeF(w*2, self.length))

    def paint(self, painter: QPainter, style: QStyleOptionGraphicsItem, widget: QWidget):
        painter.save()
        painter.setPen(self.__current_pen)
        # painter.setRenderHint(QPainter.Antialiasing)
        w = self.__current_pen.widthF()
        if self.is_horizontal:
            painter.drawLine(w/2, 0.0, self.length-w/2, 0.0)
        else:
            painter.drawLine(0.0, w/2, 0.0, self.length-w/2)
        painter.restore()

    def hoverEnterEvent(self, event: QGraphicsSceneHoverEvent):
        self.__current_pen = self.hover_pen
        self.update()
        return True

    def hoverLeaveEvent(self, event: QGraphicsSceneHoverEvent):
        self.__current_pen = self.pen
        self.update()
        return True

    def mouseMoveEvent(self, event: QGraphicsSceneMouseEvent):
        if self.is_horizontal:
            new = self.__chart.mapToValue(event.scenePos()).y()
        else:
            new = self.__chart.mapToValue(event.scenePos()).x()
        self.value = new
        # if the value has been changed by the user
        if self.__callback is not None:
            self.__callback()

    def mousePressEvent(self, event): pass

    def mouseReleaseEvent(self, event): pass

    @property
    def is_horizontal(self) -> bool:
        return self.__is_horizontal

    @is_horizontal.setter
    def is_horizontal(self, value: bool):
        if isinstance(value, bool):
            self.__is_horizontal = value
        else:
            raise TypeError(value)

    @property
    def length(self) -> float:
        if self.is_horizontal:
            return self.__chart.plotArea().width()
        else:
            return self.__chart.plotArea().height()

    @property
    def value_range(self) -> Tuple[float, float]:
        top_left = self.__chart.mapToValue(self.__chart.plotArea().topLeft())
        bottom_right = self.__chart.mapToValue(
            self.__chart.plotArea().bottomRight())
        if self.is_horizontal:
            left, right = top_left.y(), bottom_right.y()
        else:
            left, right = top_left.x(), bottom_right.x()
        if left < right:
            return left, right
        else:
            return right, left

    @property
    def value(self) -> float:
        return self.__value

    @value.setter
    def value(self, chart_value: Union[int, float]):
        left, right = self.value_range
        if chart_value < left or chart_value > right:
            return
        if self.is_horizontal:
            pos = self.__chart.mapToPosition(QPointF(0.0, chart_value))
            self.setY(pos.y())
        else:
            pos = self.__chart.mapToPosition(QPointF(chart_value, 0.0))
            self.setX(pos.x())
        self.__value = chart_value

    @property
    def pen(self) -> QPen:
        return self.__pen

    @pen.setter
    def pen(self, value: QPen):
        if isinstance(value, QPen):
            self.__pen = value
        else:
            raise TypeError(value)

    @property
    def hover_pen(self) -> QPen:
        return self.__hover_pen

    @hover_pen.setter
    def hover_pen(self, value: QPen):
        if isinstance(value, QPen):
            self.__hover_pen = value
        else:
            raise TypeError(value)

    def connect_to_chart(self, chart: QtCharts.QChart):
        self.__chart = chart
        chart.scene().addItem(self)
        chart.plotAreaChanged.connect(self.on_plot_area_changed)
        self.on_plot_area_changed(chart.plotArea())

    def disconnect_from_chart(self):
        self.__chart.scene().removeItem(self)
        self.__chart.plotAreaChanged.disconnect(self.on_plot_area_changed)

    def on_plot_area_changed(self, plot_area: QRectF):
        if self.is_horizontal:
            self.setX(plot_area.left())
        else:
            self.setY(plot_area.top())
        self.value = self.value



__all__ = ["TaskWindow"]

from enum import Enum, unique
from math import sqrt
from typing import Iterable, Tuple

from PySide2.QtCore import Qt
from PySide2.QtWidgets import QGridLayout, QLabel, QMainWindow, QWidget

from QGrain.models.FittingResult import FittingResult
from QGrain.resolvers.HeadlessResolver import FittingTask


@unique
class ProcessState(Enum):
    Unknown = 0
    Succeeded = 1
    Failed = -1


class TaskWindow(QMainWindow):
    WIDTH_WEIGHT = 16
    HEIGHT_WEIGHT = 9

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.init_ui()
        self.task_state_labels = {}

    def init_ui(self):
        self.central_widget = QWidget(self)
        self.setCentralWidget(self.central_widget)
        self.main_layout = QGridLayout(self.central_widget)

        self.state_labels_widget = QWidget()
        self.main_layout.addWidget(self.state_labels_widget, 0, 0)
        self.state_labels_layout = QGridLayout(self.state_labels_widget)


        self.setWindowTitle(self.tr("The States of Fitting Tasks"))
        # self.setWindowFlags(Qt.Drawer | Qt.WindowStaysOnTopHint)
        self.setWindowFlags(Qt.Drawer)

    def change_label_state(self, label: QLabel, state: ProcessState):
        if state == ProcessState.Unknown:
            label.setStyleSheet("QLabel{background:#FFE140;}")
        elif state == ProcessState.Succeeded:
            label.setStyleSheet("QLabel{background:#00A779;}")
        elif state == ProcessState.Failed:
            label.setStyleSheet("QLabel{background:#ED002F;}")
        else:
            raise NotImplementedError(state)

    def on_task_initialized(self, tasks: Iterable[FittingTask]):
        for i, label in self.task_state_labels.items():
            self.state_labels_layout.removeWidget(label)
        self.task_state_labels.clear()

        task_number = len(tasks)
        # row number = base_number*HEIGHT_WEIGHT
        # column number = base_number*WIDTH_WEIGHT
        base_number = sqrt(task_number/(self.WIDTH_WEIGHT*self.HEIGHT_WEIGHT))
        column_number = base_number*self.WIDTH_WEIGHT // 1 + 1

        for i, task in enumerate(tasks):
            label = QLabel()
            label.setFixedSize(16, 16)
            self.change_label_state(label, ProcessState.Unknown)
            label.setToolTip(task.sample.name)
            row = i // column_number
            col = i % column_number
            self.state_labels_layout.addWidget(label, row, col)
            self.task_state_labels.update({task.sample.uuid: label})

    def on_task_state_updated(self, states: Iterable[Tuple]):
        self.show()
        for sample_id, succeeded in states:
            if succeeded:
                label = self.task_state_labels[sample_id]
                self.change_label_state(label, ProcessState.Succeeded)

    def on_task_finished(self, succeeded_results: Iterable[FittingResult], failed_tasks: Iterable[FittingTask]):
        for task in failed_tasks:
            label = self.task_state_labels[task.sample_id]
            self.change_label_state(label, ProcessState.Failed)



__all__ = ["Resolver"]

from enum import Enum, unique
from typing import Dict, Iterable, List, Tuple

import numpy as np
from scipy.optimize import OptimizeResult, basinhopping, minimize

from QGrain.algorithms import AlgorithmData, DistributionType
from QGrain.models.FittingResult import FittingResult
from QGrain.models.SampleData import SampleData


class Resolver:
    """
    The base class of resolvers.
    """
    def __init__(self, global_optimization_maxiter=100,
                 global_optimization_success_iter=3,
                 global_optimization_stepsize=1.0,
                 final_tolerance=1e-100,
                 final_maxiter=1000,
                 minimizer_tolerance=1e-8,
                 minimizer_maxiter=500):
        self.__distribution_type = DistributionType.GeneralWeibull
        self.__component_number = 3
        self.refresh()

        # algorithms settings
        self.__global_optimization_maxiter = global_optimization_maxiter
        self.__global_optimization_success_iter = global_optimization_success_iter
        self.__global_optimization_stepsize = global_optimization_stepsize
        self.__minimizer_tolerance = minimizer_tolerance
        self.__minimizer_maxiter = minimizer_maxiter
        self.__final_tolerance = final_tolerance
        self.__final_maxiter = final_maxiter

        # the related data of current sample
        self.sample_name = None # type: str
        self.real_x = None # type: np.ndarray
        self.x_offset = 0.0 # type: float
        self.bin_numbers = None # type: np.ndarray
        self.fitting_space_x = None # type: np.ndarray
        self.target_y = None # type: np.ndarray
        # parameters to preprocess the data
        self.start_index = None # type: int
        self.end_index = None # type: int
        self.fitting_history = None # type: List[np.ndarray]

    @property
    def distribution_type(self) -> DistributionType:
        return self.__distribution_type

    @distribution_type.setter
    def distribution_type(self, value: DistributionType):
        if type(value) != DistributionType:
            return
        self.__distribution_type = value
        self.refresh()

    @property
    def component_number(self) -> int:
        return self.__component_number

    @component_number.setter
    def component_number(self, value: int):
        if type(value) != int:
            return
        if value < 1:
            return
        self.__component_number = value
        self.refresh()

    def refresh(self):
        self.algorithm_data = AlgorithmData.get_algorithm_data(self.distribution_type, self.component_number)
        self.initial_guess = self.algorithm_data.defaults

    @staticmethod
    def get_squared_sum_of_residual_errors(
            values: np.ndarray, targets: np.ndarray) -> float:
        errors = np.sum(np.square(values - targets))
        return errors

    @staticmethod
    def get_valid_data_range(target_y: np.ndarray, slice_data: bool=True):
        start_index = 0
        end_index = len(target_y)
        if slice_data:
            for i, value in enumerate(target_y):
                if value > 0.0:
                    if i == 0:
                        break
                    else:
                        start_index = i-1
                        break
            # search from tail to head
            for i, value in enumerate(target_y[start_index+1:][::-1]):
                if value > 0.0:
                    if i <= 1:
                        break
                    else:
                        end_index = (i-1)*(-1)
                        break
        return start_index, end_index

    # hooks
    def on_data_fed(self, sample_name: str):
        pass

    def on_data_not_prepared(self):
        pass

    def on_fitting_started(self):
        self.fitting_history = []

    def on_fitting_finished(self):
        pass

    def on_global_fitting_failed(self, algorithm_result: OptimizeResult):
        pass

    def on_global_fitting_succeeded(self, algorithm_result: OptimizeResult):
        pass

    def on_final_fitting_failed(self, algorithm_result: OptimizeResult):
        pass

    def on_exception_raised_while_fitting(self, exception: Exception):
        pass

    def local_iteration_callback(self, fitted_params: Iterable[float]):
        self.fitting_history.append(fitted_params)

    def global_iteration_callback(self, fitted_params: Iterable[float], function_value: float, accept: bool):
        pass

    def on_fitting_succeeded(self, algorithm_result: OptimizeResult):
        pass

    def preprocess_data(self):
        self.start_index, self.end_index = Resolver.get_valid_data_range(self.target_y)
        # Normal and General Weibull distribution need to use x offset to get better performance.
        # Because if do not do that, the search space will be larger,
        # and hence increse the difficulty of searching.
        if self.distribution_type == DistributionType.Normal or \
                self.distribution_type == DistributionType.GeneralWeibull:
            self.x_offset = self.start_index
        else:
            self.x_offset = 0.0
        # Bin numbers are similar to log(x), but will not raise negative value.
        # This will make the fitting of some distributions (e.g. Weibull) easier.
        self.bin_numbers = np.array(range(len(self.target_y)), dtype=np.float64) + 1
        # fitting under the bin numbers' space
        if self.distribution_type == DistributionType.Normal or \
                self.distribution_type == DistributionType.Weibull or \
                self.distribution_type == DistributionType.GeneralWeibull:
            self.fitting_space_x = self.bin_numbers
        else:
            raise NotImplementedError(self.distribution_type)

    def feed_data(self, sample: SampleData):
        self.sample_name = sample.name
        self.real_x = sample.classes
        self.target_y = sample.distribution
        self.preprocess_data()
        self.on_data_fed(sample.name)

    @property
    def data_prepared(self) -> bool:
        if self.real_x is None:
            # all these attributes should be `None`
            # otherwise the codes are incorrect
            assert self.sample_name is None
            assert self.target_y is None
            assert self.fitting_space_x is None
            assert self.bin_numbers is None
            assert self.start_index is None
            assert self.end_index is None
            return False
        else:
            return True

    def change_settings(self, **kwargs):
        for key, value in kwargs.items():
            if key == "global_optimization_maxiter":
                self.__global_optimization_maxiter = value
            elif key == "global_optimization_success_iter":
                self.__global_optimization_success_iter = value
            elif key == "global_optimization_stepsize":
                self.__global_optimization_stepsize = value
            elif key == "minimizer_tolerance":
                self.__minimizer_tolerance = value
            elif key == "minimizer_maxiter":
                self.__minimizer_maxiter = value
            elif key == "final_tolerance":
                self.__final_tolerance = value
            elif key == "final_maxiter":
                self.__final_maxiter = value
            else:
                raise NotImplementedError(key)

    def get_fitting_result(self, fitted_params: Iterable[float],
                           fitting_history: List[np.ndarray] = None):
        result = FittingResult(self.sample_name, self.real_x,
                               self.fitting_space_x, self.bin_numbers,
                               self.target_y, self.algorithm_data,
                               fitted_params, self.x_offset,
                               fitting_history=self.fitting_history if fitting_history is None else fitting_history)
        return result

    def try_fit(self):
        if not self.data_prepared:
            self.on_data_not_prepared()
            return
        self.on_fitting_started()

        def closure(args):
            # using partial values (i.e. don't use unnecessary zero values)
            # will highly improve the performance of algorithms
            x_to_fit = self.fitting_space_x[self.start_index: self.end_index]-self.x_offset
            y_to_fit = self.target_y[self.start_index: self.end_index]
            current_values = self.algorithm_data.mixed_func(x_to_fit, *args)
            return Resolver.get_squared_sum_of_residual_errors(current_values, y_to_fit)*100

        minimizer_kwargs = dict(method="SLSQP",
                                bounds=self.algorithm_data.bounds,
                                constraints=self.algorithm_data.constrains,
                                callback=self.local_iteration_callback,
                                options={"maxiter": self.__minimizer_maxiter,
                                         "ftol": self.__minimizer_tolerance})
        try:
            global_algorithm_result = \
                basinhopping(closure, x0=self.initial_guess,
                             minimizer_kwargs=minimizer_kwargs,
                             callback=self.global_iteration_callback,
                             niter_success=self.__global_optimization_success_iter,
                             niter=self.__global_optimization_maxiter,
                             stepsize=self.__global_optimization_stepsize)

            if global_algorithm_result.lowest_optimization_result.success or \
                    global_algorithm_result.lowest_optimization_result.status == 9:
                self.on_global_fitting_succeeded(global_algorithm_result)
            else:
                self.on_global_fitting_failed(global_algorithm_result)
                self.on_fitting_finished()
                return
            final_algorithm_result = \
                minimize(closure, method="SLSQP",
                         x0=global_algorithm_result.x,
                         bounds=self.algorithm_data.bounds,
                         constraints=self.algorithm_data.constrains,
                         callback=self.local_iteration_callback,
                         options={"maxiter": self.__final_maxiter,
                                  "ftol": self.__final_tolerance})
            # judge if the final fitting succeed
            # see https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.fmin_slsqp.html
            if final_algorithm_result.success or final_algorithm_result.status == 9:
                self.on_fitting_succeeded(final_algorithm_result)
                self.on_fitting_finished()
                return
            else:
                self.on_final_fitting_failed(final_algorithm_result)
                self.on_fitting_finished()
                return
        except Exception as e:
            self.on_exception_raised_while_fitting(e)
            self.on_fitting_finished()



__all__ = ["ChartExportingDialog"]

import os
import sys
from typing import Optional, Union

import numpy as np
from PySide2.QtCore import QSettings, QStandardPaths, Qt
from PySide2.QtGui import QDoubleValidator, QIntValidator
from PySide2.QtWidgets import (QDialog, QFileDialog, QGridLayout, QLabel,
                               QLineEdit, QPushButton)

QGRAIN_ROOT_PATH = os.path.dirname(os.path.dirname(__file__))

class ChartExportingDialog(QDialog):
    def __init__(self, canvas, setting_group: str):
        super().__init__(parent=canvas)
        self.canvas = canvas
        self.settings = QSettings(os.path.join(QGRAIN_ROOT_PATH, "settings", "chart_exporting.ini"), QSettings.Format.IniFormat)
        self.settings.beginGroup(setting_group)
        self.file_dialog = QFileDialog(self)
        self.init_ui()
        self.setAttribute(Qt.WA_StyledBackground, True)

    def init_ui(self):
        self.main_layout = QGridLayout(self)
        self.preview_label = QLabel()
        self.preview_label.setFixedHeight(400)
        self.preview_label.setAlignment(Qt.AlignCenter)
        self.main_layout.addWidget(self.preview_label, 0, 0, 1, 4)
        self.int_validator = QIntValidator()
        self.int_validator.setRange(100, 10000)
        self.double_validator = QDoubleValidator()
        self.double_validator.setRange(1.00, 10.00, 2)
        self.width_label = QLabel(self.tr("Width"))
        self.main_layout.addWidget(self.width_label, 1, 0)
        self.width_edit = QLineEdit()
        self.width_edit.setValidator(self.int_validator)
        self.main_layout.addWidget(self.width_edit, 1, 1)
        self.height_label = QLabel(self.tr("Height"))
        self.main_layout.addWidget(self.height_label, 1, 2)
        self.height_edit = QLineEdit()
        self.height_edit.setValidator(self.int_validator)
        self.main_layout.addWidget(self.height_edit, 1, 3)
        self.pixel_ratio_label = QLabel(self.tr("Pixel Ratio"))
        self.main_layout.addWidget(self.pixel_ratio_label, 2, 0)
        self.pixel_ratio_edit = QLineEdit()
        self.pixel_ratio_edit.setValidator(self.double_validator)
        self.main_layout.addWidget(self.pixel_ratio_edit, 2, 1)
        self.format_options = {self.tr("Windows Bitmap (*.bmp)"): "BMP",
                               self.tr("Joint Photographic Experts Group (*.jpg)"): "JPG",
                               self.tr("Portable Network Graphics (*.png)"): "PNG",
                               self.tr("Portable Pixmap (*.ppm)"): "PPM",
                               self.tr("X11 Bitmap (*.xbm)"): "XBM",
                               self.tr("X11 Pixmap (*.xpm)"): "XPM",
                               self.tr("Scalable Vector Graphics (*.svg)"): "SVG"}
        self.support_formats = [name for description, name in self.format_options.items()]
        self.cancel_button = QPushButton(self.tr("Cancel"))
        self.main_layout.addWidget(self.cancel_button, 3, 0, 1, 2)
        self.export_button = QPushButton(self.tr("Export"))
        self.main_layout.addWidget(self.export_button, 3, 2, 1, 2)

        self.width_edit.textChanged.connect(self.update)
        self.height_edit.textChanged.connect(self.update)
        self.pixel_ratio_edit.textChanged.connect(self.update)

        self.cancel_button.clicked.connect(self.close)
        self.export_button.clicked.connect(lambda: self.save_figure())

        self.restore()

    def restore(self):
        width = self.settings.value("width", defaultValue="800")
        height = self.settings.value("height", defaultValue="600")
        pixel_ratio = self.settings.value("pixel_ratio", defaultValue="1.0")
        # must end group before setting texts
        # because it will modify the settings
        self.width_edit.setText(width)
        self.height_edit.setText(height)
        self.pixel_ratio_edit.setText(pixel_ratio)

    @property
    def width(self) -> int:
        return self.settings.value("width", defaultValue=800, type=int)

    @width.setter
    def width(self, value: Union[int, str]):
        if isinstance(value, str) and value != "":
            value = int(value)
        if isinstance(value, int) and 100 <= value <= 10000:
            self.settings.setValue("width", value)

    @property
    def height(self) -> int:
        return self.settings.value("height", defaultValue=600, type=int)

    @height.setter
    def height(self, value: Union[int, str]):
        if isinstance(value, str) and value != "":
            value = int(value)
        if isinstance(value, int) and 100 <= value <= 10000:
            self.settings.setValue("height", value)

    @property
    def pixel_ratio(self) -> float:
        return self.settings.value("pixel_ratio", defaultValue=1.0, type=float)

    @pixel_ratio.setter
    def pixel_ratio(self, value: Union[int, float, str]):
        if isinstance(value, str) and value != "":
            value = float(value)
        if isinstance(value, (int, float)) and 1.0 <= value <= 100.0:
            self.settings.setValue("pixel_ratio", value)

    @property
    def last_format(self) -> str:
        return self.settings.value("last_format", defaultValue="png", type=str)

    @last_format.setter
    def last_format(self, value: str):
        if isinstance(value, str) and value in self.support_formats:
            self.settings.setValue("last_format", value)

    def update(self):
        # save to settings
        self.width = self.width_edit.text()
        self.height = self.height_edit.text()
        self.pixel_ratio = self.pixel_ratio_edit.text()

        pixmap = self.canvas.get_pixmap(width=self.width, height=self.height,
                                        pixel_ratio=self.pixel_ratio)

        preview_width = 400
        preview_height = preview_width * self.height / self.width
        preview_pixel_ratio = self.width * self.pixel_ratio / preview_width
        pixmap.setDevicePixelRatio(preview_pixel_ratio)
        self.preview_label.setPixmap(pixmap)

    def save_figure(self, filename: str = None):
        if filename is None:
            desktop_path = QStandardPaths.standardLocations(QStandardPaths.DesktopLocation)[0]
            filename, format_description = self.file_dialog.getSaveFileName(
                self, self.tr("Select Filename"),
                desktop_path, ";;".join([description for description, name in self.format_options.items()]))
            format_name = self.format_options[format_description]
        else:
            format_name = filename.split(".")[-1].upper()

        if format_name in ("BMP", "JPG", "PNG", "PPM", "XBM", "XPM"):
            self.canvas.export_pixmap(filename, width=self.width, height=self.height,
                                      pixel_ratio=self.pixel_ratio)
            self.last_format = format_name
        elif format_name == "SVG":
            self.canvas.export_to_svg(filename)
            self.last_format = format_name
        else:
            raise NotImplementedError(format_name)



__all__ = ["SettingWindow"]

from PySide2.QtCore import QSettings, Qt, Signal
from PySide2.QtGui import QIcon, QIntValidator, QValidator
from PySide2.QtWidgets import QGridLayout, QMainWindow, QPushButton, QWidget

from QGrain.ui.AlgorithmSetting import AlgorithmSetting
from QGrain.ui.AppSetting import AppSetting
from QGrain.ui.DataSetting import DataSetting


class SettingWindow(QMainWindow):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.init_ui()
        # self.setWindowFlags(Qt.Drawer | Qt.WindowStaysOnTopHint)
        self.setWindowFlags(Qt.Drawer)
        self.setWindowTitle(self.tr("Settings"))

    def init_ui(self):
        self.central_widget = QWidget(self)
        self.setCentralWidget(self.central_widget)
        self.main_layout = QGridLayout(self.central_widget)
        self.data_setting = DataSetting()
        self.main_layout.addWidget(self.data_setting, 0, 0, 1, 2)
        self.algorithm_setting = AlgorithmSetting()
        self.main_layout.addWidget(self.algorithm_setting, 1, 0, 1, 2)
        self.app_setting = AppSetting()
        self.main_layout.addWidget(self.app_setting, 2, 0, 1, 2)

    def setup_all(self):
        pass



__all__ = ["ControlPanel"]

import logging
import os

import numpy as np
from PySide2.QtCore import Qt, QTimer, Signal
from PySide2.QtGui import QFont
from PySide2.QtWidgets import (QCheckBox, QGridLayout, QLabel, QMessageBox,
                               QPushButton, QRadioButton, QWidget)

from QGrain.algorithms import DistributionType
from QGrain.models.FittingResult import FittingResult
from QGrain.models.SampleDataset import SampleDataset


class ControlPanel(QWidget):
    sigDistributionTypeChanged = Signal(int)
    sigComponentNumberChanged = Signal(int)
    sigFocusSampleChanged = Signal(int) # index of the sample
    sigDataSettingsChanged = Signal(dict)
    sigObserveIterationChanged = Signal(bool)
    sigInheritParamsChanged = Signal(bool)
    sigGUIResolverFittingStarted = Signal()
    sigGUIResolverFittingCanceled = Signal()
    sigMultiProcessingFittingStarted = Signal()
    logger = logging.getLogger("root.ui.ControlPanel")
    gui_logger = logging.getLogger("GUI")

    def __init__(self, parent=None, **kargs):
        super().__init__(parent, **kargs)
        self.__component_number = 2
        self.__data_index = 0
        self.sample_names = None
        self.auto_fit_flag = True
        self.auto_run_timer = QTimer()
        self.auto_run_timer.setSingleShot(True)
        self.auto_run_timer.timeout.connect(self.on_auto_run_timer_timeout)
        self.auto_run_flag = False

        self.init_ui()
        self.connect_all()

        self.msg_box = QMessageBox(self)
        self.msg_box.setWindowFlags(Qt.Drawer)

    def init_ui(self):
        self.main_layout = QGridLayout(self)
        # Distribution type
        self.distribution_type_label = QLabel(self.tr("Distribution Type:"), self)
        self.distribution_type_label.setToolTip(self.tr("Select the base distribution function of each component."))
        self.distribution_normal_radio_button = QRadioButton(self.tr("Normal"), self)
        self.distribution_normal_radio_button.setToolTip(self.tr("See [%s] for more details.") % "https://en.wikipedia.org/wiki/Normal_distribution")
        self.distribution_weibull_radio_button = QRadioButton(self.tr("Weibull"), self)
        self.distribution_weibull_radio_button.setToolTip(self.tr("See [%s] for more details.") % "https://en.wikipedia.org/wiki/Weibull_distribution")
        self.distribution_gen_weibull_radio_button = QRadioButton(self.tr("Gen. Weibull"), self)
        self.distribution_gen_weibull_radio_button.setToolTip(self.tr("General Weibull distribution which has an additional location parameter."))
        self.main_layout.addWidget(self.distribution_type_label, 0, 0, 1, 2)
        self.main_layout.addWidget(self.distribution_normal_radio_button, 0, 1)
        self.main_layout.addWidget(self.distribution_weibull_radio_button, 0, 2)
        self.main_layout.addWidget(self.distribution_gen_weibull_radio_button, 0, 3)
        # Component number
        self.component_number_label = QLabel(self.tr("Component Number:"))
        self.component_number_label.setToolTip(self.tr("Select the component number of the mixed distribution."))
        self.component_number_display = QLabel(self.tr("Unknown"))
        self.component_number_add_button = QPushButton(self.tr("+"))
        self.component_number_add_button.setToolTip(self.tr("Click to add one component.\nIt should be less than 11."))
        self.component_reduce_button = QPushButton(self.tr("-"))
        self.component_reduce_button.setToolTip(self.tr("Click to reduce one component.\nIt should be positive."))
        self.main_layout.addWidget(self.component_number_label, 1, 0)
        self.main_layout.addWidget(self.component_number_display, 1, 1)
        self.main_layout.addWidget(self.component_number_add_button, 1, 2)
        self.main_layout.addWidget(self.component_reduce_button, 1, 3)
        # Some usual settings
        self.observe_iteration_checkbox = QCheckBox(self.tr("Observe Iteration"))
        self.observe_iteration_checkbox.setToolTip(self.tr("Whether to display the iteration procedure."))
        self.inherit_params_checkbox = QCheckBox(self.tr("Inherit Parameters"))
        self.inherit_params_checkbox.setToolTip(self.tr("Whether to inherit the parameters of last fitting.\nIt will improve the accuracy and efficiency when the samples are continuous."))
        self.main_layout.addWidget(self.observe_iteration_checkbox, 2, 0, 1, 2)
        self.main_layout.addWidget(self.inherit_params_checkbox, 2, 2, 1, 2)
        self.auto_fit_checkbox = QCheckBox(self.tr("Auto Fit"))
        self.auto_fit_checkbox.setToolTip(self.tr("Whether to automaticlly fit after the sample data changed."))
        self.auto_record_checkbox = QCheckBox(self.tr("Auto Record"))
        self.auto_record_checkbox.setToolTip(self.tr("Whether to automaticlly record the fitting result after fitting finished."))
        self.main_layout.addWidget(self.auto_fit_checkbox, 3, 0, 1, 2)
        self.main_layout.addWidget(self.auto_record_checkbox, 3, 2, 1, 2)
        # Target data to fit
        self.data_index_label = QLabel(self.tr("Current Sample:"))
        self.data_index_label.setToolTip(self.tr("Current sample to fit."))
        self.data_index_display = QLabel(self.tr("Unknown"))
        self.data_index_previous_button = QPushButton(self.tr("Previous"))
        self.data_index_previous_button.setToolTip(self.tr("Click to back to the previous sample."))
        self.data_index_next_button = QPushButton(self.tr("Next"))
        self.data_index_next_button.setToolTip(self.tr("Click to jump to the next sample."))
        self.main_layout.addWidget(self.data_index_label, 4, 0)
        self.main_layout.addWidget(self.data_index_display, 4, 1)
        self.main_layout.addWidget(self.data_index_previous_button, 4, 2)
        self.main_layout.addWidget(self.data_index_next_button, 4, 3)
        # Control bottons
        self.auto_run = QPushButton(self.tr("Auto Run Orderly"))
        self.auto_run.setToolTip(self.tr("Click to run the program automatically.\nThe samples from current to the end will be processed one by one."))
        self.cancel_run = QPushButton(self.tr("Cancel"))
        self.cancel_run.setToolTip(self.tr("Click to cancel the fitting progress."))
        self.try_fit_button = QPushButton(self.tr("Try Fit"))
        self.try_fit_button.setToolTip(self.tr("Click to fit the current sample."))
        self.record_button = QPushButton(self.tr("Record"))
        self.record_button.setToolTip(self.tr("Click to record the current fitting result.\nNote: It will record the LAST SUCCESS fitting result, NOT CURRENT SAMPLE."))
        self.main_layout.addWidget(self.auto_run, 5, 0)
        self.main_layout.addWidget(self.cancel_run, 5, 1)
        self.main_layout.addWidget(self.try_fit_button, 5, 2)
        self.main_layout.addWidget(self.record_button, 5, 3)
        # Multi cores
        self.multiprocessing_button = QPushButton(self.tr("Multi Cores Fitting"))
        self.multiprocessing_button.setToolTip(self.tr("Click to fit all samples. It will utilize all cores of cpu to accelerate calculation."))
        self.main_layout.addWidget(self.multiprocessing_button, 6, 0, 1, 4)

    def connect_all(self):
        self.distribution_normal_radio_button.clicked.connect(self.on_distribution_type_changed)
        self.distribution_weibull_radio_button.clicked.connect(self.on_distribution_type_changed)
        self.distribution_gen_weibull_radio_button.clicked.connect(self.on_distribution_type_changed)
        self.component_number_add_button.clicked.connect(self.on_component_number_add_clicked)
        self.component_reduce_button.clicked.connect(self.on_component_number_reduce_clicked)

        self.data_index_previous_button.clicked.connect(self.on_data_index_previous_clicked)
        self.data_index_next_button.clicked.connect(self.on_data_index_next_clicked)

        self.observe_iteration_checkbox.stateChanged.connect(self.on_show_iteration_changed)
        self.inherit_params_checkbox.stateChanged.connect(self.on_inherit_params_changed)
        self.auto_fit_checkbox.stateChanged.connect(self.on_auto_fit_changed)
        self.auto_record_checkbox.stateChanged.connect(self.on_auto_record_changed)

        self.auto_run.clicked.connect(self.on_auto_run_clicked)
        self.cancel_run.clicked.connect(self.on_cancel_run_clicked)
        self.try_fit_button.clicked.connect(self.on_try_fit_clicked)
        self.multiprocessing_button.clicked.connect(self.on_multiprocessing_clicked)

    @property
    def component_number(self):
        return self.__component_number

    @component_number.setter
    def component_number(self, value: int):
        # check the validity
        # component number should be non-negative
        if value < 1 or value > 10:
            self.gui_logger.info(self.tr("The component number should be >= 1 and <= 10."))
            return
        # update the label to display the value
        self.component_number_display.setText(str(value))
        self.__component_number = value
        self.sigComponentNumberChanged.emit(value)
        # auto emit target data change signal again when component number changed
        if self.sample_names is not None:
            self.data_index = self.data_index

    @property
    def data_index(self) -> int:
        return self.__data_index

    @data_index.setter
    def data_index(self, value: int):
        if not self.has_data:
            return
        if value < 0 or value >= self.data_length:
            self.gui_logger.info(self.tr("It has reached the first/last sample."))
            return
        # update the label to display the name of this sample
        self.data_index_display.setText(self.sample_names[value])
        self.__data_index = value
        self.logger.debug("Data index has been set to [%d].", value)
        self.sigFocusSampleChanged.emit(value)
        if self.auto_fit_flag:
            self.sigGUIResolverFittingStarted.emit()

    @property
    def current_name(self) -> str:
        return self.sample_names[self.data_index]

    @property
    def has_data(self) -> bool:
        if self.data_length <= 0:
            return False
        else:
            return True

    @property
    def data_length(self) -> int:
        if self.sample_names is None:
            return 0
        else:
            return len(self.sample_names)

    def show_message(self, title: str, message: str):
        self.msg_box.setWindowTitle(title)
        self.msg_box.setText(message)
        self.msg_box.exec_()

    def show_info(self, message: str):
        self.show_message(self.tr("Info"), message)

    def show_warning(self, message: str):
        self.show_message(self.tr("Warning"), message)

    def show_error(self, message: str):
        self.show_message(self.tr("Error"), message)

    def check_data_loaded(self, show_msg=True):
        if not self.has_data:
            if show_msg:
                self.show_warning(self.tr("The grain size data has not been loaded, the operation is invalid."))
            return False
        else:
            return True

    def on_component_number_add_clicked(self):
        self.component_number += 1

    def on_component_number_reduce_clicked(self):
        self.component_number -= 1

    def on_distribution_type_changed(self):
        if self.distribution_normal_radio_button.isChecked():
            distribution_type = DistributionType.Normal
        elif self.distribution_weibull_radio_button.isChecked():
            distribution_type = DistributionType.Weibull
        else:
            assert self.distribution_gen_weibull_radio_button.isChecked()
            distribution_type = DistributionType.GeneralWeibull
        self.logger.info("Distribution type has been changed to [%s].", distribution_type)
        self.sigDistributionTypeChanged.emit(distribution_type.value)

        if self.sample_names is not None:
            self.data_index = self.data_index

    def on_data_index_previous_clicked(self):
        if not self.check_data_loaded():
            return
        self.data_index -= 1

    def on_data_index_next_clicked(self):
        if not self.check_data_loaded():
            return
        self.data_index += 1

    def on_show_iteration_changed(self, state: Qt.CheckState):
        if state == Qt.Checked:
            self.sigObserveIterationChanged.emit(True)
        else:
            self.sigObserveIterationChanged.emit(False)

    def on_inherit_params_changed(self, state: Qt.CheckState):
        if state == Qt.Checked:
            self.sigInheritParamsChanged.emit(True)
        else:
            self.sigInheritParamsChanged.emit(False)

    def on_auto_fit_changed(self, state: Qt.CheckState):
        if state == Qt.Checked:
            self.auto_fit_flag = True
        else:
            self.auto_fit_flag = False

    def on_auto_record_changed(self, state: Qt.CheckState):
        if state == Qt.Checked:
            self.sigDataSettingsChanged.emit({"auto_record": True})
        else:
            self.sigDataSettingsChanged.emit({"auto_record": False})

    def on_data_loaded(self, dataset: SampleDataset):
        self.sample_names = [sample.name for sample in dataset.samples]
        self.logger.info("Data was loaded.")
        self.data_index = 0
        self.logger.info("Data index has been set to 0.")

    def on_data_selected(self, index: int):
        self.data_index = index
        self.logger.debug("Sample data at [%d] is selected.", index)

    def change_enable_states(self, enable: bool):
        if self.auto_run_flag and enable:
            return
        self.distribution_weibull_radio_button.setEnabled(enable)
        self.distribution_normal_radio_button.setEnabled(enable)
        self.component_number_add_button.setEnabled(enable)
        self.component_reduce_button.setEnabled(enable)
        self.data_index_previous_button.setEnabled(enable)
        self.data_index_next_button.setEnabled(enable)
        self.observe_iteration_checkbox.setEnabled(enable)
        self.inherit_params_checkbox.setEnabled(enable)
        self.auto_fit_checkbox.setEnabled(enable)
        self.auto_record_checkbox.setEnabled(enable)
        self.auto_run.setEnabled(enable)
        self.try_fit_button.setEnabled(enable)
        self.record_button.setEnabled(enable)
        self.multiprocessing_button.setEnabled(enable)

    def on_fitting_suceeded(self, result: FittingResult):
        if self.auto_run_flag and result.has_invalid_value:
            self.logger.warning("The fitting result may be not valid, auto running stoped.")
            self.gui_logger.warning(self.tr("The fitting result may be not valid, auto running stoped."))
            self.auto_run_flag = False
            self.change_enable_states(True)

        if self.data_index == self.data_length-1:
            self.logger.info("The auto running has reached the last sample and stoped.")
            self.gui_logger.info(self.tr("The auto running has reached the last sample and stoped."))
            self.auto_run_flag = False
            self.change_enable_states(True)

        if self.auto_run_flag:
            self.auto_run_timer.start(5)

    def on_auto_run_timer_timeout(self):
        self.data_index += 1

    def on_auto_run_clicked(self):
        if not self.check_data_loaded():
            return
        self.auto_fit_checkbox.setCheckState(Qt.Checked)
        self.auto_record_checkbox.setCheckState(Qt.Checked)
        # from current sample to fit, to avoid that it need to resart from the first sample every time
        self.data_index = self.data_index
        self.auto_run_flag = True
        self.logger.info("Auto running started from sample [%s].", self.current_name)

    def on_cancel_run_clicked(self):
        if self.auto_run_flag:
            self.auto_run_flag = False
            self.change_enable_states(True)
            self.logger.info("Auto run flag has been changed to False.")

        self.sigGUIResolverFittingCanceled.emit()

    def on_try_fit_clicked(self):
        self.sigGUIResolverFittingStarted.emit()

    def on_multiprocessing_clicked(self):
        if not self.check_data_loaded():
            return
        self.sigMultiProcessingFittingStarted.emit()

    def on_fitting_started(self):
        self.change_enable_states(False)

    def on_fitting_finished(self):
        self.change_enable_states(True)

    def on_fitting_failed(self, message: str):
        if self.auto_run_flag:
            self.auto_run_flag = False
            self.change_enable_states(True)
            self.logger.info("Auto running was canceled due to the failure of fitting.")
        self.show_error(self.tr("Fitting failed. {0}").format(message))

    def setup_all(self):
        self.component_number = 3
        self.distribution_gen_weibull_radio_button.setChecked(True)
        self.observe_iteration_checkbox.setCheckState(Qt.Unchecked)
        self.inherit_params_checkbox.setCheckState(Qt.Checked)
        self.auto_fit_checkbox.setCheckState(Qt.Checked)
        self.auto_record_checkbox.setCheckState(Qt.Checked)



__all__ = ["ViewDataManager", "RecordedDataTable"]

import logging
from datetime import date, datetime, time
from typing import Dict, List, Tuple
from uuid import UUID

import numpy as np
from PySide2.QtCore import QCoreApplication, QEventLoop, QObject, Qt, Signal
from PySide2.QtGui import QCursor
from PySide2.QtWidgets import (QAbstractItemView, QGridLayout, QMenu,
                               QTableWidget, QTableWidgetItem, QWidget)

from QGrain.algorithms import DistributionType
from QGrain.models.FittingResult import FittingResult


class ViewDataManager(QObject):
    MAX_PARAM_COUNT = 3
    COMPONENT_SPAN = 1
    TABLE_HEADER_ROWS = 2
    TABLE_ROW_EXPAND_SETP = 50
    COMPONENT_START_COLUMN = 3
    def __init__(self, table:QTableWidget, is_detailed: bool):
        super().__init__()
        self.table = table
        self.is_detailed = is_detailed
        self.existing_key2display_name = {
            "fraction": self.tr("Fraction"),
            "mean": self.tr("Mean")+" (μm)",
            "median": self.tr("Median")+" (μm)",
            "mode": self.tr("Mode")+" (μm)",
            # "variance": self.tr("Variance"),
            "standard_deviation": self.tr("Standard Deviation"),
            "skewness": self.tr("Skewness"),
            "kurtosis": self.tr("Kurtosis")}
        self.summary_keys = ("fraction", "mean", "median")
        self.detailed_keys = tuple([key for key, display_name in self.existing_key2display_name.items()])
        self.records = [] #List[SingleViewData]

    @property
    def data_count(self) -> int:
        return len(self.records)

    @property
    def max_component_number(self) -> int:
        if self.data_count == 0:
            return 0
        else:
            return max([record.component_number for record in self.records])

    @property
    def component_columns(self) -> int:
        if self.is_detailed:
            return len(self.detailed_keys) + self.MAX_PARAM_COUNT
        else:
            return len(self.summary_keys)

    @property
    def actual_component_columns(self):
        return self.component_columns + self.COMPONENT_SPAN

    def get_distribution_name(self, distribution_type: DistributionType):
        if distribution_type == DistributionType.Normal:
            return self.tr("Normal")
        elif distribution_type == DistributionType.Weibull:
            return self.tr("Weibull")
        elif distribution_type == DistributionType.GeneralWeibull:
            return self.tr("Gen. Weibull")
        else:
            raise NotImplementedError(distribution_type)

    # the local func to handle write
    # also put data validation here
    def write(self, row, col, value, format_str="{0:0.2f}"):
        # the local func to check the value validation
        def check_value(value):
            accepted_types = (bool, datetime, date, time)
            if value is None:
                return "None"
            # must judge this first,
            # because `Formula` can pass `numpy.isreal` but raise error in `numpy.isnan`
            elif type(value) in accepted_types:
                return value
            # if value is real number (i.e. int, float, numpy.int32, numpy.float32, etc.)
            elif np.isreal(value):
                # this func will both recognize float("nan") and numpy.nan
                if np.isnan(value):
                    return "NaN"
                # similar to above
                elif np.isinf(value):
                    return "Inf"
                else:
                    return format_str.format(value)
            # if the type of value is unaccepted, use its readable str to write
            else:
                return str(value)

        item = QTableWidgetItem(check_value(value))
        item.setTextAlignment(Qt.AlignCenter)
        self.table.setItem(row, col, item)


    def update_headers(self):
        max_component_number = self.max_component_number
        # update headers
        self.write(0, 0, self.tr("Sample Name"))
        self.table.setSpan(0, 0, self.TABLE_HEADER_ROWS, 1)
        self.write(0, 1, self.tr("Distribution Type"))
        self.table.setSpan(0, 1, self.TABLE_HEADER_ROWS, 1)
        self.write(0, 2, self.tr("Mean Squared Error"))
        self.table.setSpan(0, 2, self.TABLE_HEADER_ROWS, 1)
        for comp_index in range(max_component_number):
            first_column_index = comp_index*self.actual_component_columns+self.COMPONENT_START_COLUMN
            self.write(0, first_column_index, self.tr("Component {0}").format(comp_index+1))
            self.table.setSpan(0, first_column_index, 1, self.component_columns)
            if self.is_detailed:
                headers = list((self.existing_key2display_name[key] for key in self.detailed_keys))
                for i in range(self.MAX_PARAM_COUNT):
                    headers.append(self.tr("Parameter {0}").format(i+1))
            else:
                headers = list((self.existing_key2display_name[key] for key in self.summary_keys))
            for i, header in enumerate(headers):
                self.write(1, first_column_index+i, header)



    def add_records(self, records_to_add: List[FittingResult]):
        if records_to_add is None or len(records_to_add) == 0:
            return
        length = len(records_to_add)
        # if it's necessary to expand the rows and columns of table
        least_row_number = self.data_count + length + self.TABLE_HEADER_ROWS
        if  least_row_number>= self.table.rowCount():
            self.table.setRowCount(least_row_number + self.TABLE_ROW_EXPAND_SETP)
        max_component_number_before = self.max_component_number
        max_component_number_new = max([record.component_number for record in records_to_add])
        # reset the column number of table
        column_count = max(max_component_number_before, max_component_number_new) * self.actual_component_columns + self.COMPONENT_START_COLUMN
        self.table.setColumnCount(column_count)

        # update contents
        first_row_index = self.data_count + self.TABLE_HEADER_ROWS
        for record_index, record in enumerate(records_to_add):
            # will bring thread unsafe issue
            # QCoreApplication.processEvents(QEventLoop.ExcludeUserInputEvents)
            row = first_row_index + record_index
            self.write(row, 0, record.name)
            self.write(row, 1, self.get_distribution_name(record.distribution_type))
            self.write(row, 2, record.mean_squared_error, format_str="{0:0.2E}")
            if self.is_detailed:
                keys = self.detailed_keys
            else:
                keys = self.summary_keys
            for i, component in enumerate(record.components):
                last_column = 0
                for key_index, key in enumerate(keys):
                    column_index = i*self.actual_component_columns + self.COMPONENT_START_COLUMN + key_index
                    last_column = column_index
                    self.write(row, column_index, component.__getattribute__(key))
                if self.is_detailed:
                    for param_index, param_value in enumerate(component.params):
                        self.write(row, last_column + param_index + 1, param_value)
            # store the records
            self.records.append(record)

        # means there is greater component number and the headers are need to be updated
        if max_component_number_new > max_component_number_before:
            self.update_headers()
            self.table.resizeColumnsToContents()


    def remove_selected_records(self) -> List[Tuple[UUID, str]]:
        # get rows to remove
        rows_to_remove = set()
        # The behaviour of `selectedRanges` differs when clicking in table or at the edge
        # When clicking in table it returns multi ranges (each row)
        # When clicking at the edge it returns a single range
        for item in self.table.selectedRanges():
            for i in range(item.topRow(), min(self.data_count+self.TABLE_HEADER_ROWS, item.bottomRow()+1)):
                # do not remove headers
                if i >= self.TABLE_HEADER_ROWS:
                    rows_to_remove.add(i)
        rows_to_remove = list(rows_to_remove)
        rows_to_remove.sort()

        uuids_and_names_to_remove = []
        offset = 0
        for row in rows_to_remove:
            self.table.removeRow(row-offset)
            removed_view_data = self.records.pop(row-offset-self.TABLE_HEADER_ROWS)
            uuids_and_names_to_remove.append((removed_view_data.uuid, removed_view_data.name))
            offset += 1

        # reset the column number of table
        column_count = self.max_component_number * self.actual_component_columns + self.COMPONENT_START_COLUMN
        self.table.setColumnCount(column_count)

        return uuids_and_names_to_remove

    def get_selected_record(self) -> FittingResult:
        ranges = self.table.selectedRanges()
        if len(ranges) != 0:
            row = ranges[0].topRow()
            return self.records[row-self.TABLE_HEADER_ROWS]


class RecordedDataTable(QWidget):
    logger = logging.getLogger("root.ui.RecordedDataTable")
    gui_logger = logging.getLogger("GUI")
    sigRemoveRecords = Signal(list)
    sigShowDistribution = Signal(FittingResult)
    sigShowLoss = Signal(FittingResult)
    sigGenerateDistributionVideo = Signal(FittingResult)
    def __init__(self, parent=None, **kargs):
        super().__init__(parent, **kargs)
        self.init_ui()

    def init_ui(self):
        self.main_layout = QGridLayout(self)
        self.table = QTableWidget()
        self.table.setEditTriggers(QAbstractItemView.NoEditTriggers)
        self.table.setSelectionBehavior(QAbstractItemView.SelectRows)
        self.table.setAlternatingRowColors(True)
        self.table.setContextMenuPolicy(Qt.CustomContextMenu)
        self.main_layout.addWidget(self.table)
        self.main_layout.setContentsMargins(0, 0, 0, 0)
        self.manager = ViewDataManager(self.table, is_detailed=True)
        self.menu = QMenu(self.table)
        self.remove_action = self.menu.addAction(self.tr("Remove"))
        self.remove_action.triggered.connect(self.remove_selection)
        self.show_distribution_action = self.menu.addAction(self.tr("Show Distribution"))
        self.show_distribution_action.triggered.connect(self.show_distribution)
        self.generate_distribution_video_action = self.menu.addAction(self.tr("Generate Distribution Video"))
        self.generate_distribution_video_action.triggered.connect(self.generate_distribution_video)
        self.show_loss_action = self.menu.addAction(self.tr("Show Loss"))
        self.show_loss_action.triggered.connect(self.show_loss)
        self.table.customContextMenuRequested.connect(self.show_menu)

    def show_menu(self, pos):
        self.menu.popup(QCursor.pos())

    def on_data_recorded(self, results: List[FittingResult]):
        self.manager.add_records(results)
        self.logger.debug("Recorded data has been updated in table widget.")

    def remove_selection(self):
        uuids_and_names = self.manager.remove_selected_records()
        self.logger.debug("The selected records below will be removed: [%s].", uuids_and_names)
        self.sigRemoveRecords.emit(uuids_and_names)

    def show_distribution(self):
        record = self.manager.get_selected_record()
        if record is not None:
            self.sigShowDistribution.emit(record)

    def show_loss(self):
        record = self.manager.get_selected_record()
        if record is not None:
            self.sigShowLoss.emit(record)

    def generate_distribution_video(self):
        record = self.manager.get_selected_record()
        if record is not None:
            self.sigGenerateDistributionVideo.emit(record)



__all__ = ["Canvas"]

import math
import sys
from typing import Optional, Union

import numpy as np
from PySide2.QtCharts import QtCharts
from PySide2.QtCore import QPointF, QRectF, QSizeF, Qt, QTimer
from PySide2.QtGui import (QBrush, QColor, QDoubleValidator, QFont,
                           QIntValidator, QPainter, QPen, QPixmap)
from PySide2.QtSvg import QSvgGenerator
from PySide2.QtWidgets import (QAction, QApplication, QGraphicsItem,
                               QGraphicsScene, QGraphicsSceneDragDropEvent,
                               QGraphicsSceneHoverEvent,
                               QGraphicsSceneMouseEvent, QGraphicsView,
                               QGridLayout, QStyleOptionGraphicsItem, QWidget)

from QGrain.ui.ChartExportingDialog import ChartExportingDialog


class Canvas(QWidget):
    def __init__(self, setting_group: str, parent: Optional[QWidget] = None):
        super().__init__(parent=parent)
        self.setAttribute(Qt.WA_StyledBackground, True)
        self.main_layout = QGridLayout(self)
        self.main_layout.setContentsMargins(0, 0, 0, 0)
        self.chart = QtCharts.QChart()
        self.chart_view = QtCharts.QChartView()
        self.chart_view.setChart(self.chart)
        self.main_layout.addWidget(self.chart_view)

        self.export_dialog = ChartExportingDialog(self, setting_group)
        self.setContextMenuPolicy(Qt.ActionsContextMenu)
        # self.edit_action = QAction(self.tr("Edit"))
        # self.addAction(self.edit_action)
        self.export_action = QAction(self.tr("Export"))
        self.export_action.triggered.connect(self.on_export_clicked)
        self.addAction(self.export_action)


    def setup_chart_style(self):
        self.chart.setTitleFont(QFont("Times New Roman", 12))
        # self.chart.setTitleBrush(QBrush(QColor(0xff0000))) # set color
        self.chart.axisX().setTitleFont(QFont("Times New Roman", 10))
        # self.chart.axisX().setTitleBrush(QBrush(QColor(0xff0000)))
        self.chart.axisY().setTitleFont(QFont("Times New Roman", 10))
        # self.chart.axisX().setTitleBrush(QBrush(QColor(0xff0000)))
        self.chart.legend().setAlignment(Qt.AlignTop)
        self.chart.legend().setMarkerShape(QtCharts.QLegend.MarkerShapeFromSeries)
        self.chart.legend().setFont(QFont("Times New Roman", 10))
        # self.chart.setAnimationOptions(QtCharts.QChart.AllAnimations)
        self.chart.setBackgroundVisible(False)
        self.chart_view.setRenderHint(QPainter.Antialiasing)

    def set_theme_mode(self, is_dark: bool):
        if is_dark:
            self.chart.setTheme(QtCharts.QChart.ChartThemeDark)
        else:
            self.chart.setTheme(QtCharts.QChart.ChartThemeLight)

    def to_points(self, x: np.ndarray, y: np.ndarray):
        return [QPointF(x_value, y_value) for x_value, y_value in zip(x, y)]

    def show_demo(self, axis_x: QtCharts.QAbstractAxis,
                 axis_y: QtCharts.QAbstractAxis,
                 x_log=False, y_log=False):
        def love(x, a):
            return np.abs(x)**(2/3) + (0.9*np.sqrt(np.abs(3.3-x**2))) * np.sin(a*np.pi*x)
        series = QtCharts.QLineSeries()
        series.setName(self.tr("Demo"))
        a = 3.3
        x = np.linspace(-np.sqrt(3.3), np.sqrt(3.3), 1000)
        y = love(x, a)

        series.replace(self.to_points(10**x if x_log else x, 10**y if y_log else y))
        self.chart.addSeries(series)
        series.attachAxis(axis_x)
        series.attachAxis(axis_y)
        scale = 1.2
        minX = -np.sqrt(3.3) * scale
        maxX = np.sqrt(3.3) * scale
        minY = -1.5737869944381024 * scale
        maxY = 2.367369351208529 * scale
        if x_log:
            axis_x.setRange(10**minX, 10**maxX)
        else:
            axis_x.setRange(minX, maxX)
        if y_log:
            axis_y.setRange(10**minY, 10**maxY)
        else:
            axis_y.setRange(minY, maxY)

        def update():
            nonlocal a
            a += 0.01
            if a > 33:
                a = 3.3
            y = love(x, a)
            series.replace(self.to_points(10**x if x_log else x, 10**y if y_log else y))
        self.demo_series = series
        self.demo_timer = QTimer()
        self.demo_timer.timeout.connect(update)
        self.demo_timer.start(1000/60)

    def stop_demo(self):
        if hasattr(self, "demo_timer"):
            self.demo_timer.stop()
            del self.demo_timer
        if hasattr(self, "demo_series"):
            self.chart.removeSeries(self.demo_series)
            del self.demo_series

    def get_pixmap(self, width: Union[int] = 800, height: Union[int] = 600,
                   pixel_ratio: Union[int, float] = 1.0):
        geometry = self.chart_view.saveGeometry()
        assert isinstance(width, int)
        assert width > 0
        assert isinstance(height, int)
        assert height > 0
        assert isinstance(pixel_ratio, (int, float))
        # set geometry to make it in order
        self.chart_view.setGeometry(0, 0, width, height)
        image = QPixmap(int(width*pixel_ratio),
                       int(height*pixel_ratio))
        image.setDevicePixelRatio(pixel_ratio)
        self.chart_view.render(image)
        self.chart_view.restoreGeometry(geometry)
        return image

    def export_pixmap(self, filename: str,
                      width: Union[int] = 800,
                      height: Union[int] = 600,
                      pixel_ratio: Union[int, float] = 1.0):
        pixmap = self.get_pixmap(width=width, height=height, pixel_ratio=pixel_ratio)
        pixmap.save(filename)

    def export_svg(self, filename: str):
        geometry = self.chart_view.saveGeometry()
        # set geometry to make it in order
        self.chart_view.setGeometry(0, 0, 800, 600)
        generator = QSvgGenerator()
        generator.setFileName(filename)
        generator.setTitle("Generated by QGrain (version 0.2.7)")
        generator.setDescription("""This svg image was generated by QGrain, you can take secondary process to make it can be published.""")
        self.chart_view.render(generator)
        self.chart_view.restoreGeometry(geometry)

    def on_export_clicked(self):
        self.export_dialog.update()
        self.export_dialog.exec_()


if __name__ == "__main__":
    app = QApplication(sys.argv)
    canvs = Canvas("Canvas")
    canvs.chart.legend().hide()
    axisX = QtCharts.QValueAxis()
    axisY = QtCharts.QValueAxis()
    canvs.chart.addAxis(axisX, Qt.AlignBottom)
    canvs.chart.addAxis(axisY, Qt.AlignLeft)
    canvs.show()
    canvs.show_demo(axisX, axisY)
    # canvs.stopDemo()
    sys.exit(app.exec_())



__all__ = ["column_to_char", "to_cell_name", "DataWriter"]

import csv
import json
import os
from datetime import date, datetime, time
from typing import List

import numpy as np
import xlsxwriter
import xlwt

from QGrain.algorithms import DistributionType
from QGrain.models.DataLoader import FileType
from QGrain.models.FittingResult import FittingResult

QGRAIN_ROOT_PATH = os.path.dirname(os.path.dirname(__file__))

def column_to_char(column_index: int):
    column = column_index + 1
    column_str = str()
    while column != 0:
        res = column % 26
        if res == 0:
            res = 26
            column -= 26
        column_str = chr(ord('A') + res - 1) + column_str
        column = column // 26
    return column_str

def to_cell_name(row: int, column: int):
    return "{0}{1}".format(column_to_char(column), row+1)


class DataWriter:
    """
    The class to save the fitting results to local files.
    """
    MAX_PARAM_COUNT = 3
    def __init__(self):
        super().__init__()
        self.style_file_path = os.path.join(QGRAIN_ROOT_PATH, "settings", "chart_styles.json")
        # see https://xlsxwriter.readthedocs.io/chart.html

    def get_distribution_name(self, distribution_type: DistributionType):
        if distribution_type == DistributionType.Normal:
            return "Normal"
        elif distribution_type == DistributionType.Weibull:
            return "Weibull"
        elif distribution_type == DistributionType.GeneralWeibull:
            return "Gen. Weibull"
        else:
            raise NotImplementedError(distribution_type)

    def try_save_data(self, filename: str, file_type: FileType,
                      results: List[FittingResult], draw_charts: bool):
        if filename is None or filename == "":
            raise ValueError(filename)

        if file_type == FileType.XLSX:
            self.try_save_as_excel(filename, results, True, draw_charts)
        elif file_type == FileType.XLS:
            self.try_save_as_excel(filename, results, False, draw_charts)
        elif file_type == FileType.CSV:
            self.try_save_as_csv(filename, results)
        else:
            raise NotImplementedError(file_type)

    # If type is csv, it will use built-in csv module to handle this request.
    def try_save_as_csv(self, filename: str, results: List[FittingResult]):
        # use `newline=""` to avoid redundant newlines
        with open(filename, "w", newline="", encoding="utf-8") as f:
            w = csv.writer(f)
            max_component_number = max([result.component_number for result in results])
            # write the hearders
            headers = ["Sample Name", "Distribution Type", "Mean Squared Error",
            "Pearson Correlation Coefficient", "P Value", "Kendall's TAU", "P Value",
            "Spearman Correlation Coefficient", "P Value"]
            for i in range(max_component_number):
                headers.append("")
                headers.append("Fraction (%)")
                headers.append("Mean (mu m)")
                headers.append("Median (mu m)")
                headers.append("Mode (mu m)")
                headers.append("Variance")
                headers.append("Standard Deviation")
                headers.append("Skewness")
                headers.append("Kurtosis")
                for j in range(self.MAX_PARAM_COUNT):
                    headers.append("Parameter" + " {0}".format(j+1))
            w.writerow(headers)
            # write the contents
            for result in results:
                row = [result.name,
                       self.get_distribution_name(result.distribution_type),
                       result.mean_squared_error, *result.pearson_r,
                       *result.kendall_tau, *result.spearman_r]
                for component in result.components:
                    row.append("")
                    row.append(component.fraction * 100.0)
                    row.append(component.mean)
                    row.append(component.median)
                    row.append(component.mode)
                    row.append(component.variance)
                    row.append(component.standard_deviation)
                    row.append(component.skewness)
                    row.append(component.kurtosis)
                    # write the parameters of each base distribution
                    for i in range(self.MAX_PARAM_COUNT):
                        if i < result.param_count:
                            row.append(component.params[i])
                        else:
                            row.append("")
                w.writerow(row)

    def try_save_as_excel(self, filename: str, results: List[FittingResult],
                          is_xlsx: bool, draw_charts: bool):
        # the local func to check the value validation
        # must block unaccepted values before calling the `write` func of worksheet
        def check_value(value):
            accepted_types = (bool, datetime, date, time, xlwt.Formula)
            if value is None:
                return "None"
            # must judge this first,
            # because `Formula` can pass `numpy.isreal` but raise error in `numpy.isnan`
            elif type(value) in accepted_types:
                return value
            # if value is real number (i.e. int, float, numpy.int32, numpy.float32, etc.)
            elif np.isreal(value):
                # this func will both recognize float("nan") and numpy.nan
                if np.isnan(value):
                    return "NaN"
                # similar to above
                elif np.isinf(value):
                    return "Inf"
                else:
                    return value
            # if the type of value is unaccepted, use its readable str to write
            else:
                return str(value)

        def write(sheet, row, col, value, style):
            sheet.write(row, col, check_value(value), style)

        # the wrappers to handle the differences between xlwt and xlsxwriter
        if is_xlsx:
            # see https://xlsxwriter.readthedocs.io/worksheet.html#merge_range
            def mwrite(sheet, lrow, lcol, rrow, rcol, value, style):
                sheet.merge_range(lrow, lcol, rrow, rcol, check_value(value), style)
            def set_col(sheet, col, width):
                sheet.set_column(col, col, width)
        else:
            def mwrite(sheet, lrow, lcol, rrow, rcol, value, style):
                sheet.write_merge(lrow, rrow, lcol, rcol, check_value(value), style)
            def set_col(sheet, col, width):
                sheet.col(col).width = width*256
        # generate different styles
        if is_xlsx:
            book = xlsxwriter.Workbook(filename=filename)
            # see https://xlsxwriter.readthedocs.io/format.html#format
            global_style = book.add_format({"font_name": "Times New Roman", "font_size": 12, "align": "center", "valign": "vcenter", "pattern": 1, "fg_color": "#F5F5F5", "num_format": "0.00"})
            dark_global_style = book.add_format({"font_name": "Times New Roman", "font_size": 12, "align": "center", "valign": "vcenter", "pattern": 1, "fg_color": "#A9A9A9", "num_format": "0.00"})
            detail_global_style = book.add_format({"font_name": "Times New Roman", "font_size": 12, "align": "center", "valign": "vcenter", "pattern": 1, "fg_color": "#F5F5F5", "num_format": "0.0000"})
            dark_detail_global_style = book.add_format({"font_name": "Times New Roman", "font_size": 12, "align": "center", "valign": "vcenter", "pattern": 1, "fg_color": "#A9A9A9", "num_format": "0.0000"})
            exp_style = book.add_format({"font_name": "Times New Roman", "font_size": 12, "align": "center", "valign": "vcenter", "pattern": 1, "fg_color": "#F5F5F5", "num_format": "0.00E+00"})
            dark_exp_style = book.add_format({"font_name": "Times New Roman", "font_size": 12, "align": "center", "valign": "vcenter", "pattern": 1, "fg_color": "#A9A9A9", "num_format": "0.00E+00"})
            fraction_style = book.add_format({"font_name": "Times New Roman", "font_size": 12, "align": "center", "valign": "vcenter", "pattern": 1, "fg_color": "#F5F5F5", "num_format": "0.00%"})
            dark_fraction_style = book.add_format({"font_name": "Times New Roman", "font_size": 12, "align": "center", "valign": "vcenter", "pattern": 1, "fg_color": "#A9A9A9", "num_format": "0.00%"})
            header_style = book.add_format({"font_name": "Times New Roman", "font_size": 12, "font_color": "#FFF5EE", "bold": True, "align": "center", "valign": "vcenter", "text_wrap": True, "pattern": 1, "fg_color": "#4169E1", "num_format": "0.00"})
            # use this sheet to record the statistic data
            summary_sheet_name = "Summary" # this name will be referred in detail sheet
            summary_sheet = book.add_worksheet(summary_sheet_name)
            detail_sheet_name = "Detail"
            detail_sheet = book.add_worksheet(detail_sheet_name)
        else:
            book = xlwt.Workbook()
            # create styles by easyxf
            # see https://github.com/python-excel/xlwt/blob/master/examples/xlwt_easyxf_simple_demo.py
            global_style = xlwt.easyxf("font: name Times New Roman, height 240; align: vert centre, horiz center; pattern: pattern solid, fore_colour 1", num_format_str="0.00")
            dark_global_style = xlwt.easyxf("font: name Times New Roman, height 240; align: vert centre, horiz center; pattern: pattern solid, fore_colour 22", num_format_str="0.00")
            detail_global_style = xlwt.easyxf("font: name Times New Roman, height 240; align: vert centre, horiz center; pattern: pattern solid, fore_colour 1", num_format_str="0.0000")
            dark_detail_global_style = xlwt.easyxf("font: name Times New Roman, height 240; align: vert centre, horiz center; pattern: pattern solid, fore_colour 22", num_format_str="0.0000")
            exp_style = xlwt.easyxf("font: name Times New Roman, height 240; align: vert centre, horiz center; pattern: pattern solid, fore_colour 1", num_format_str="0.00E+00")
            dark_exp_style = xlwt.easyxf("font: name Times New Roman, height 240; align: vert centre, horiz center; pattern: pattern solid, fore_colour 22", num_format_str="0.00E+00")
            fraction_style = xlwt.easyxf("font: name Times New Roman, height 240; align: vert centre, horiz center; pattern: pattern solid, fore_colour 1", num_format_str="0.00%")
            dark_fraction_style = xlwt.easyxf("font: name Times New Roman, height 240; align: vert centre, horiz center; pattern: pattern solid, fore_colour 22", num_format_str="0.00%")
            header_style = xlwt.easyxf("font: name Times New Roman, height 320, bold on, colour 1; align: wrap on, vert centre, horiz center; pattern: pattern solid, fore_colour 30", num_format_str="0.00")
            # use this sheet to record the statistic data
            summary_sheet_name = "Summary" # this name will be referred in detailed sheet
            summary_sheet = book.add_sheet(summary_sheet_name)
            detail_sheet_name = "Detail"
            detail_sheet = book.add_sheet(detail_sheet_name)

        max_component_number = max([result.component_number for result in results])
        SUMMARY_HEADER_ROWS = 2
        SUMMARY_COMPONENTS = 8 + self.MAX_PARAM_COUNT
        SUMMARY_COLUMN_SPAN = SUMMARY_COMPONENTS + 1
        SUMMARY_COMPONENT_START_COLUMN = 10
        # sheet 1.
        # headers of summary sheet
        mwrite(summary_sheet, 0, 0, 1, 0, "Sample Name", header_style)
        mwrite(summary_sheet, 0, 1, 1, 1, "Distribution Type", header_style)
        mwrite(summary_sheet, 0, 2, 1, 2, "Mean Squared Error", header_style)
        mwrite(summary_sheet, 0, 3, 0, 4, "Pearson's Correlation Coefficient", header_style)
        write(summary_sheet, 1, 3, "Correlation", header_style)
        write(summary_sheet, 1, 4, "Two-tailed P-value", header_style)
        mwrite(summary_sheet, 0, 5, 0, 6, "Kendall’s TAU", header_style)
        write(summary_sheet, 1, 5, "The TAU Statistic", header_style)
        write(summary_sheet, 1, 6, "Two-sided P-value", header_style)
        mwrite(summary_sheet, 0, 7, 0, 8, "Spearman's Correlation Coefficient", header_style)
        write(summary_sheet, 1, 7, "Correlation", header_style)
        write(summary_sheet, 1, 8, "Two-sided P-value", header_style)
        for i in range(max_component_number):
            left_col = i*SUMMARY_COLUMN_SPAN+SUMMARY_COMPONENT_START_COLUMN
            mwrite(summary_sheet, 0, left_col, 0, left_col+SUMMARY_COMPONENTS-1, "Component {0}".format(i+1), header_style)
            write(summary_sheet, 1, left_col, "Fraction", header_style)
            write(summary_sheet, 1, left_col+1, "Mean (μm)", header_style)
            write(summary_sheet, 1, left_col+2, "Median (μm)", header_style)
            write(summary_sheet, 1, left_col+3, "Mode (μm)", header_style)
            write(summary_sheet, 1, left_col+4, "Variance", header_style)
            write(summary_sheet, 1, left_col+5, "Standard Deviation", header_style)
            write(summary_sheet, 1, left_col+6, "Skewness", header_style)
            write(summary_sheet, 1, left_col+7, "Kurtosis", header_style)
            for j in range(self.MAX_PARAM_COUNT):
                write(summary_sheet, 1, left_col+8+j, "Parameter" + " {0}".format(j+1), header_style)

        # set widths of columns
        set_col(summary_sheet, 0, 16)
        set_col(summary_sheet, 1, 16)
        for i in range(2, 9):
            if i % 2 == 0:
                set_col(summary_sheet, i, 12)
            else:
                set_col(summary_sheet, i, 8)
        for i in range(SUMMARY_COMPONENT_START_COLUMN, max_component_number*SUMMARY_COLUMN_SPAN+SUMMARY_COMPONENT_START_COLUMN):
            set_col(summary_sheet, i, 8)
        # contents of summary sheet
        for row, result in enumerate(results, SUMMARY_HEADER_ROWS):
            # take this way to let color changes alternately
            if row % 2 == 0:
                current_global_style = global_style
                current_exp_style = exp_style
                current_fraction_style = fraction_style
            else:
                current_global_style = dark_global_style
                current_exp_style = dark_exp_style
                current_fraction_style = dark_fraction_style
            write(summary_sheet, row, 0, result.name, current_global_style)
            write(summary_sheet, row, 1, self.get_distribution_name(result.distribution_type), current_global_style)
            write(summary_sheet, row, 2, result.mean_squared_error, current_exp_style)
            write(summary_sheet, row, 3, result.pearson_r[0], current_global_style)
            write(summary_sheet, row, 4, result.pearson_r[1], current_exp_style)
            write(summary_sheet, row, 5, result.kendall_tau[0], current_global_style)
            write(summary_sheet, row, 6, result.kendall_tau[1], current_exp_style)
            write(summary_sheet, row, 7, result.spearman_r[0], current_global_style)
            write(summary_sheet, row, 8, result.spearman_r[1], current_exp_style)

            for i, component in enumerate(result.components):
                left_col = i*SUMMARY_COLUMN_SPAN+SUMMARY_COMPONENT_START_COLUMN
                write(summary_sheet, row, left_col, component.fraction, current_fraction_style)
                write(summary_sheet, row, left_col+1, component.mean, current_global_style)
                write(summary_sheet, row, left_col+2, component.median, current_global_style)
                write(summary_sheet, row, left_col+3, component.median, current_global_style)
                write(summary_sheet, row, left_col+4, component.variance, current_global_style)
                write(summary_sheet, row, left_col+5, component.standard_deviation, current_global_style)
                write(summary_sheet, row, left_col+6, component.skewness, current_global_style)
                write(summary_sheet, row, left_col+7, component.kurtosis, current_global_style)

                for j in range(self.MAX_PARAM_COUNT):
                    if j < result.param_count:
                        write(summary_sheet, row, left_col+8+j, component.params[j], current_global_style)

        # sheet 2.
        # headers of detail sheet
        DETAIL_COMPONENT_START_COLUMN = 3
        write(detail_sheet, 0, 0, "Sample Name", header_style)
        write(detail_sheet, 0, 1, "Series Name", header_style)
        set_col(detail_sheet, 0, 16)
        set_col(detail_sheet, 1, 16)
        classes_length = len(results[0].real_x)
        for i in range(DETAIL_COMPONENT_START_COLUMN, DETAIL_COMPONENT_START_COLUMN+classes_length):
            set_col(detail_sheet, i, 12)
        # classes headers
        for col, value in enumerate(results[0].real_x, DETAIL_COMPONENT_START_COLUMN):
            write(detail_sheet, 0, col, value, header_style)
        # contents of detail sheet
        row = 1
        for data_index, result in enumerate(results):
            if data_index % 2 == 0:
                current_global_style = detail_global_style
            else:
                current_global_style = dark_detail_global_style
            mwrite(detail_sheet, row, 0, row+result.component_number+1, 0, result.name, current_global_style)
            # target row
            for col, value in enumerate(result.target_y, DETAIL_COMPONENT_START_COLUMN):
                write(detail_sheet, row, col, value, current_global_style)
            write(detail_sheet, row, 1, "Target", current_global_style)
            row += 1
            # sum row
            for col, value in enumerate(result.fitted_y, DETAIL_COMPONENT_START_COLUMN):
                write(detail_sheet, row, col, value, current_global_style)
            write(detail_sheet, row, 1, "Fitted Sum", current_global_style)
            row += 1
            # components' rows
            for component_index, component in enumerate(result.components):
                write(detail_sheet, row, 1, "C{0}".format(component_index+1), current_global_style)
                for col, value in enumerate(component.component_y, DETAIL_COMPONENT_START_COLUMN):
                    write(detail_sheet, row, col, value, current_global_style)
                row += 1

        # Save file if it is xls or no need to draw charts
        if not is_xlsx or not draw_charts:
            if is_xlsx:
                book.close()
                return
            else:
                book.save(filename)
                return

        # add charts to workbook
        chart_styles = None
        column_chart_number = 5
        x_margin = 10
        y_margin = 10
        width = 480
        height = 288
        try:
            with open(self.style_file_path, "r", encoding="utf-8") as style_file:
                chart_styles = json.load(style_file)
                column_chart_number = chart_styles["column_chart_number"]
                x_margin = chart_styles["x_margin"]
                y_margin = chart_styles["y_margin"]
                width = chart_styles["size"]["x_scale"]*480
                height = chart_styles["size"]["y_scale"]*288
        except OSError:
            pass

        chart_sheet = book.add_worksheet("Charts")
        row = 1
        for i, result in enumerate(results):
            chart = book.add_chart({'type': 'scatter'})
            if chart_styles is not None:
                chart.set_x_axis(chart_styles["x_axis"])
                chart.set_y_axis(chart_styles["y_axis"])
                chart.set_title({**chart_styles["title"], "name": [detail_sheet_name, row, 0]})
                chart.set_legend(chart_styles["legend"])
                chart.set_size(chart_styles["size"])
                for j in range(result.component_number+2):
                    if j  == 0:
                        style = chart_styles["target_series"]
                    elif j == 1:
                        style = chart_styles["fitted_series"]
                    else:
                        style = chart_styles["component_series"]
                    chart.add_series({**{"name": [detail_sheet_name, row, 1],
                                        "categories": [detail_sheet_name, 0, DETAIL_COMPONENT_START_COLUMN, 0, DETAIL_COMPONENT_START_COLUMN+classes_length-1],
                                        "values": [detail_sheet_name, row, DETAIL_COMPONENT_START_COLUMN, row, DETAIL_COMPONENT_START_COLUMN+classes_length-1]}, **style})
                    row += 1
            else:
                chart.set_x_axis({"log_base": 10, "crossing": 0.01})
                chart.set_title({"name": [detail_sheet_name, row, 0]})
                for j in range(result.component_number+2):
                    chart.add_series({"name": [detail_sheet_name, row, 1],
                                      "categories": [detail_sheet_name, 0, DETAIL_COMPONENT_START_COLUMN, 0, DETAIL_COMPONENT_START_COLUMN+classes_length-1],
                                      "values": [detail_sheet_name, row, DETAIL_COMPONENT_START_COLUMN, row, DETAIL_COMPONENT_START_COLUMN+classes_length-1]})
                    row += 1

            x_offset = (i%column_chart_number) * (width+x_margin)
            y_offset = (i//column_chart_number) * (height+y_margin)
            chart_sheet.insert_chart('A1', chart, {"x_offset": x_offset, "y_offset": y_offset})

        book.close()



__all__ = ["LossCanvas"]

import logging
import sys

from PySide2.QtCharts import QtCharts
from PySide2.QtCore import Qt

from QGrain.models.FittingResult import FittingResult
from QGrain.ui.Canvas import Canvas


class LossCanvas(Canvas):
    logger = logging.getLogger("root.ui.LossCanvas")
    gui_logger = logging.getLogger("GUI")

    def __init__(self, parent=None, is_dark=True):
        super().__init__("LossCanvas", parent=parent)
        self.set_theme_mode(is_dark)
        self.init_chart()
        self.setup_chart_style()
        self.chart.legend().hide()
        self.observe_iteration_tag = False

    def init_chart(self):
        # init axes
        self.axis_x = QtCharts.QValueAxis()
        self.axis_x.setTickInterval(20)
        self.chart.addAxis(self.axis_x, Qt.AlignBottom)
        self.axis_y = QtCharts.QLogValueAxis()
        self.axis_y.setBase(10.0)
        self.axis_y.setLabelFormat("%.1e")
        self.chart.addAxis(self.axis_y, Qt.AlignLeft)
        # init the series
        self.loss_series = QtCharts.QLineSeries()
        self.loss_series.setName(self.tr("Loss"))
        self.chart.addSeries(self.loss_series)
        # attach series to axes
        self.loss_series.attachAxis(self.axis_x)
        self.loss_series.attachAxis(self.axis_y)
        # set title
        self.chart.setTitle(self.tr("Loss Canvas"))
        # set labels
        self.axis_x.setTitleText(self.tr("Iteration"))
        self.axis_y.setTitleText(self.tr("Loss"))

        self.show_demo(self.axis_x, self.axis_y, y_log=True)

    def on_observe_iteration_changed(self, value: bool):
        self.observe_iteration_tag = value

    def show_fitting_result(self, result: FittingResult):
        if not self.observe_iteration_tag:
            return
        self.stop_demo()
        self.loss_series.clear()
        max_loss = -sys.maxsize
        min_loss = sys.maxsize
        for i, result_in_process in enumerate(result.history):
            self.loss_series.append(i, result_in_process.mean_squared_error)
            if result_in_process.mean_squared_error > max_loss:
                max_loss = result_in_process.mean_squared_error
            if result_in_process.mean_squared_error < min_loss:
                min_loss = result_in_process.mean_squared_error
            # self.chart.setTitle(("{0} "+self.tr("Iteration")+" ({1})").format(result.name, i))
            # self.axis_x.setRange(0.0, current_iteration)
            # self.axis_y.setRange(min_Loss, max_loss)

        self.axis_x.setRange(0, result.iteration_number-1)
        self.axis_y.setRange(min_loss, max_loss)
        self.chart.setTitle(result.name)


if __name__ == "__main__":
    import sys
    from PySide2.QtWidgets import QApplication
    app = QApplication(sys.argv)
    canvas = LossCanvas(is_dark=False)
    canvas.chart.legend().hide()
    canvas.show()
    sys.exit(app.exec_())



__all__ = ["QGRAIN_VERSION", "main"]

import logging
import os
import shutil
import sys
from logging.handlers import RotatingFileHandler, TimedRotatingFileHandler
from multiprocessing import freeze_support

from PySide2.QtCore import QSettings, QTranslator
from PySide2.QtGui import QIcon, QPixmap
from PySide2.QtWidgets import QApplication, QSplashScreen

from QGrain.ui.MainWindow import GUILogHandler, MainWindow

QGRAIN_VERSION = "0.2.7.1"
QGRAIN_ROOT_PATH = os.path.dirname(__file__)


def getdirsize(dir):
   size = 0
   for root, _, files in os.walk(dir):
      size += sum([os.path.getsize(os.path.join(root, name)) for name in files])
   return size

def create_necessary_folders():
    necessary_folders = (os.path.join(QGRAIN_ROOT_PATH, "logs"),)
    for folder in necessary_folders:
        if not os.path.exists(folder):
            os.mkdir(folder)

def get_language():
    settings = QSettings(os.path.join(QGRAIN_ROOT_PATH, "settings", "QGrain.ini"), QSettings.Format.IniFormat)
    settings.beginGroup("app")
    lang = settings.value("language", defaultValue="en", type=str)
    settings.endGroup()
    return lang

def get_theme():
    settings = QSettings(os.path.join(QGRAIN_ROOT_PATH, "settings", "QGrain.ini"), QSettings.Format.IniFormat)
    settings.beginGroup("app")
    theme = settings.value("theme", defaultValue="MaterialDark", type=str)
    settings.endGroup()
    return theme

def setup_language(app: QApplication):
    lang = get_language()
    trans = QTranslator(app)
    trans.load(os.path.join(QGRAIN_ROOT_PATH, "i18n", lang))
    app.installTranslator(trans)

def setup_theme(app: QApplication) -> bool:
    theme = get_theme()
    with open(os.path.join(QGRAIN_ROOT_PATH, "settings", "qss", "{0}.qss".format(theme))) as template:
        template_styles = template.read()
    with open(os.path.join(QGRAIN_ROOT_PATH, "settings", "custom.qss")) as custom:
        custom_style = custom.read()
    app.setStyleSheet(template_styles+custom_style)

    if theme == "Aqua":
        return False
    elif theme == "Ubuntu":
        return False
    elif theme == "ElegantDark":
        return True
    elif theme == "MaterialDark":
        return True
    else:
        raise NotImplementedError(theme)

def setup_logging(main_window: MainWindow):
    format_str = "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    file_handler = TimedRotatingFileHandler(os.path.join(QGRAIN_ROOT_PATH, "logs", "qgrain.log"), when="D", backupCount=8, encoding="utf-8")
    file_handler.setLevel(logging.DEBUG)
    file_handler.setFormatter(logging.Formatter(format_str))
    gui_handler = GUILogHandler(main_window)
    gui_handler.setLevel(logging.INFO)
    logging.basicConfig(level=logging.DEBUG, format=format_str)
    logging.getLogger().addHandler(file_handler)
    logging.getLogger("GUI").addHandler(gui_handler)

def exec_qgrain():
    create_necessary_folders()
    app = QApplication(sys.argv)
    logo = QPixmap(os.path.join(QGRAIN_ROOT_PATH, "settings", "icons", "splash_logo.png"))
    splash = QSplashScreen(logo)
    splash.show()
    app.processEvents()
    setup_language(app)
    is_dark = setup_theme(app)
    main_window = MainWindow(is_dark=is_dark)
    main_window.setWindowTitle("QGrain")
    main_window.setWindowIcon(QIcon(os.path.join(QGRAIN_ROOT_PATH, "settings", "icons", "icon.png")))
    setup_logging(main_window)
    main_window.show()
    main_window.setup_all()
    splash.finish(main_window)
    sys.exit(app.exec_())

def main():
    import multiprocessing
    multiprocessing.set_start_method('spawn', True)
    freeze_support()
    exec_qgrain()

if __name__ == "__main__":
    main()
